# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2013-2014
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2013-2014
# - Martin Barisits, <martin.barisits@cern.ch>, 2014

from traceback import format_exc

from alembic import command
from alembic.config import Config

from sqlalchemy.engine import reflection
from sqlalchemy.schema import MetaData, Table, DropTable, ForeignKeyConstraint, DropConstraint

from rucio.common.config import config_get
from rucio.core.account_counter import create_counters_for_new_account
from rucio.db.sqla import session, models
from rucio.db.sqla.constants import AccountStatus, AccountType, IdentityType


def build_database(echo=True, tests=False):
    """ Applies the schema to the database. Run this command once to build the database. """
    engine = session.get_engine(echo=echo)
    models.register_models(engine)

    # Put the database under version control
    alembic_cfg = Config(config_get('alembic', 'cfg'))
    command.stamp(alembic_cfg, "head")


def dump_schema():
    """ Creates a schema dump to a specific database. """
    engine = session.get_dump_engine()
    models.register_models(engine)


def destroy_database(echo=True):
    """ Removes the schema from the database. Only useful for test cases or malicious intents. """
    engine = session.get_engine(echo=echo)
    models.unregister_models(engine)


def drop_everything(echo=True):
    """ Pre-gather all named constraints and table names, and drop everything. This is better than using metadata.reflect();
        metadata.drop_all() as it handles cyclical constraints between tables.
        Ref. http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DropEverything
    """
    engine = session.get_engine(echo=echo)
    conn = engine.connect()

    # the transaction only applies if the DB supports
    # transactional DDL, i.e. Postgresql, MS SQL Server
    trans = conn.begin()

    inspector = reflection.Inspector.from_engine(engine)

    # gather all data first before dropping anything.
    # some DBs lock after things have been dropped in
    # a transaction.
    metadata = MetaData()

    tbs = []
    all_fks = []

    for table_name in inspector.get_table_names():
        fks = []
        for fk in inspector.get_foreign_keys(table_name):
            if not fk['name']:
                continue
            fks.append(ForeignKeyConstraint((), (), name=fk['name']))
        t = Table(table_name, metadata, *fks)
        tbs.append(t)
        all_fks.extend(fks)

    for fkc in all_fks:
        try:
            print str(DropConstraint(fkc)) + ';'
            conn.execute(DropConstraint(fkc))
        except:
            print format_exc()

    for table in tbs:
        try:
            print str(DropTable(table)).strip() + ';'
            conn.execute(DropTable(table))
        except:
            print format_exc()

    trans.commit()


def create_root_account():
    """ Inserts the default root account to an existing database. Make sure to change the default password later. """

    up_id = 'ddmlab'
    up_pwd = '2ccee6f6dd1bc2269cddd7cd5e47578e98e430539807c36df23fab7dd13e7583'
    up_email = 'ph-adp-ddm-lab@cern.ch'
    x509_id = '/C=CH/ST=Geneva/O=CERN/OU=PH-ADP-CO/CN=DDMLAB Client Certificate/emailAddress=ph-adp-ddm-lab@cern.ch'
    x509_email = 'ph-adp-ddm-lab@cern.ch'
    gss_id = 'ddmlab@CERN.CH'
    gss_email = 'ph-adp-ddm-lab@cern.ch'

    try:
        up_id = config_get('bootstrap', 'userpass_identity')
        up_pwd = config_get('bootstrap', 'userpass_pwd')
        up_email = config_get('bootstrap', 'userpass_email')
        x509_id = config_get('bootstrap', 'x509_identity')
        x509_email = config_get('bootstrap', 'x509_email')
        gss_id = config_get('bootstrap', 'gss_identity')
        gss_email = config_get('bootstrap', 'gss_email')
    except:
        pass
        # print 'Config values are missing (check rucio.cfg{.template}). Using hardcoded defaults.'

    s = session.get_session()

    account = models.Account(account='root', account_type=AccountType.SERVICE, status=AccountStatus.ACTIVE)

    identity1 = models.Identity(identity=up_id, identity_type=IdentityType.USERPASS, password=up_pwd, salt='0', email=up_email)
    iaa1 = models.IdentityAccountAssociation(identity=identity1.identity, identity_type=identity1.identity_type, account=account.account, is_default=True)

    # X509 authentication
    identity2 = models.Identity(identity=x509_id, identity_type=IdentityType.X509, email=x509_email)
    iaa2 = models.IdentityAccountAssociation(identity=identity2.identity, identity_type=identity2.identity_type, account=account.account, is_default=True)

    # GSS authentication
    identity3 = models.Identity(identity=gss_id, identity_type=IdentityType.GSS, email=gss_email)
    iaa3 = models.IdentityAccountAssociation(identity=identity3.identity, identity_type=identity3.identity_type, account=account.account, is_default=True)

    # Account counters
    create_counters_for_new_account(account='root', session=s)

    # Apply
    s.add_all([account, identity1, identity2, identity3])
    s.commit()
    s.add_all([iaa1, iaa2, iaa3])
    s.commit()



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Luis Rodrigues, <luis.rodrigues@cern.ch>, 2013


# client2.add_replica(rse_src, scope_name, did_name, did_bytes)
# client2.add_replication_rule(dids=[{'scope': scope_name, 'name': did_name}], account=account_name, copies=2, rse_expression='%s|%s'%(rse1, rse2), grouping='NONE')



from StringIO import StringIO
from datetime import datetime
from nose.tools import eq_
from nose.tools import ok_
from nose.tools import raises
from rucio.common import config
from rucio.common import dumper
from rucio.tests.common import make_temp_file
from rucio.tests.common import stubbed
from rucio.tests.common import mock_open
from rucio.tests.mock import gfal2
import __builtin__
import bz2file
import os
import requests
import tempfile

DATE_SECONDS = "2015-03-10 14:00:35"
DATE_TENTHS = "2015-03-10T14:00:35.5"
DATE_MILLISECONDS = "2015-03-10T14:00:35.5"


@raises(dumper.HTTPDownloadFailed)
def test_http_download_failed_exception_with_no_semantic_errors():
    raise dumper.HTTPDownloadFailed('some message', 500)


@raises(SystemExit)
def test_error_ends_the_program():
    dumper.error('message', 2)


def test_cacert_config_returns_a_string():
    ok_(isinstance(dumper.cacert_config(config, '.'), str))


def test_cacert_config_returns_false_if_no_cert_configured():
    with stubbed(config.config_get, lambda _, __: ''):
        eq_(dumper.cacert_config(config, '.'), False)


def test_smart_open_for_text_file():
    tmp = make_temp_file('/tmp', 'abcdef')
    ok_(isinstance(dumper.smart_open(tmp), __builtin__.file))
    os.unlink(tmp)


def test_smart_open_for_bz2_file():
    fd, path = tempfile.mkstemp()
    comp = bz2file.BZ2Compressor()
    with os.fdopen(fd, 'w') as f:
        f.write(comp.compress('abcdef'))
        f.write(comp.flush())

    ok_(not isinstance(dumper.smart_open(path), bz2file.BZ2File))
    os.unlink(path)


def test_temp_file_with_final_name_creates_a_tmp_file_and_then_removes_it():
    final_name = tempfile.mktemp()
    with dumper.temp_file('/tmp', final_name) as (tmpf, tmp_path):
        tmp_path = os.path.join('/tmp', tmp_path)
        ok_(os.path.exists(tmp_path), tmp_path)
        ok_(not os.path.exists(final_name), tmp_path)

    ok_(os.path.exists(final_name), final_name)
    ok_(not os.path.exists(tmp_path), tmp_path)
    os.unlink(final_name)


def test_temp_file_with_final_name_creates_a_tmp_file_and_keeps_it():
    with dumper.temp_file('/tmp') as (tmpf, tmp_path):
        tmp_path = os.path.join('/tmp', tmp_path)
        ok_(os.path.exists(tmp_path), tmp_path)

    ok_(os.path.exists(tmp_path), tmp_path)
    os.unlink(tmp_path)


def test_temp_file_cleanup_on_exception():
    try:
        with dumper.temp_file('/tmp') as (tmpf, tmp_path):
            tmp_path = os.path.join('/tmp', tmp_path)
            raise Exception
    except:
        pass
    finally:
        ok_(not os.path.exists(tmp_path), tmp_path)


def test_temp_file_cleanup_on_exception_with_final_name():
    final_name = tempfile.mktemp()
    try:
        with dumper.temp_file('/tmp', final_name) as (tmpf, tmp_path):
            tmp_path = os.path.join('/tmp', tmp_path)
            raise Exception
    except:
        pass
    finally:
        ok_(not os.path.exists(tmp_path), tmp_path)
        ok_(not os.path.exists(final_name), final_name)


def test_to_date_format():
    ok_(isinstance(dumper.to_datetime(DATE_SECONDS), datetime))
    ok_(isinstance(dumper.to_datetime(DATE_TENTHS), datetime))
    ok_(isinstance(dumper.to_datetime(DATE_MILLISECONDS), datetime))


def test_agis_endpoints_data_parses_proper_json():
    response = requests.Response()
    response._content = '{"x": "y"}'
    response.status_code = 200

    with stubbed(requests.get, lambda _: response):
        eq_(dumper.agis_endpoints_data(cache=False), {'x': 'y'})


def test_ddmendpoint_url_builds_url_from_agis_records():
    agisdata = [{
        'name': 'SOMEENDPOINT',
        'se': 'srm://example.com/',
        'endpoint': 'atlasdatadisk/',
    }]
    with stubbed(dumper.agis_endpoints_data, lambda: agisdata):
        eq_(dumper.ddmendpoint_url('SOMEENDPOINT'), 'srm://example.com/atlasdatadisk/')


@raises(StopIteration)
def test_ddmendpoint_url_fails_on_unexistent_entry():
    with stubbed(dumper.agis_endpoints_data, lambda: []):
        dumper.ddmendpoint_url('SOMEENDPOINT')


def test_http_download_to_file_without_session_uses_requests_get():
    response = requests.Response()
    response.status_code = 200
    response.iter_content = lambda _: ['content']

    stringio = StringIO()

    with stubbed(requests.get, lambda _, stream=False: response):
        dumper.http_download_to_file('http://example.com', stringio)

    stringio.seek(0)
    eq_(stringio.read(), 'content')


def test_http_download_to_file_with_session():
    response = requests.Response()
    response.status_code = 200
    response.iter_content = lambda _: ['content']

    stringio = StringIO()
    session = requests.Session()
    session.get = lambda _: response

    dumper.http_download_to_file('http://example.com', stringio, session)
    stringio.seek(0)
    eq_(stringio.read(), 'content')


@raises(dumper.HTTPDownloadFailed)
def test_http_download_to_file_throws_exception_on_error():
    response = requests.Response()
    response.status_code = 404
    response.iter_content = lambda _: ['content']

    stringio = StringIO()
    session = requests.Session()
    session.get = lambda _: response

    dumper.http_download_to_file('http://example.com', stringio, session)


def test_http_download_creates_file_with_content():
    response = requests.Response()
    response.status_code = 200
    response.iter_content = lambda _: ['abc']
    stringio = StringIO()

    with stubbed(requests.get, lambda _, **kw: response):
        with mock_open(dumper, stringio):
            dumper.http_download('http://example.com', 'filename')

    stringio.seek(0)
    eq_(stringio.read(), 'abc')


def test_srm_download_to_file():
    srm_file = StringIO()
    local_file = StringIO()
    srm_file.write('content')
    srm_file.seek(0)

    with gfal2.mocked_gfal2(dumper, files={'srm://example.com/file': srm_file}):
        dumper.srm_download_to_file('srm://example.com/file', local_file)

    local_file.seek(0)
    eq_(local_file.read(), 'content')


def test_srm_download_creates_file_with_content():
    srm_file = StringIO()
    local_file = StringIO()
    srm_file.write('content')
    srm_file.seek(0)

    with gfal2.mocked_gfal2(dumper, files={'srm://example.com/file': srm_file}):
        with mock_open(dumper, local_file):
            dumper.srm_download('srm://example.com/file', 'filename')

    local_file.seek(0)
    eq_(local_file.read(), 'content')



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2012-2016
# - Thomas Beermann, <thomas.beermann@cern.ch>, 2012
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2014

from os import remove

from nose.tools import raises

from rucio.client.baseclient import BaseClient
from rucio.client.client import Client
from rucio.common.config import config_get
from rucio.common.utils import get_tmp_dir
from rucio.common.exception import CannotAuthenticate, ClientProtocolNotSupported


class TestBaseClient():

    def setup(self):
        self.cacert = config_get('test', 'cacert')
        self.usercert = config_get('test', 'usercert')
        try:
            remove(get_tmp_dir() + '/.rucio_root/auth_token_root')
        except OSError, e:
            if e.args[0] != 2:
                raise e

    def tearDown(self):
        pass

    def testUserpass(self):
        """ CLIENTS (BASECLIENT): authenticate with userpass."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        BaseClient(account='root', ca_cert=self.cacert, auth_type='userpass', creds=creds)

    @raises(CannotAuthenticate)
    def testUserpassWrongCreds(self):
        """ CLIENTS (BASECLIENT): try to authenticate with wrong username."""
        creds = {'username': 'wrong', 'password': 'secret'}
        BaseClient(account='root', ca_cert=self.cacert, auth_type='userpass', creds=creds)

    @raises(CannotAuthenticate)
    def testUserpassNoCACert(self):
        """ CLIENTS (BASECLIENT): authenticate with userpass without ca cert."""
        creds = {'username': 'wrong', 'password': 'secret'}
        BaseClient(account='root', auth_type='userpass', creds=creds)

    def testx509(self):
        """ CLIENTS (BASECLIENT): authenticate with x509."""
        creds = {'client_cert': self.usercert}
        BaseClient(account='root', ca_cert=self.cacert, auth_type='x509', creds=creds)

    @raises(CannotAuthenticate)
    def testx509NonExistingCert(self):
        """ CLIENTS (BASECLIENT): authenticate with x509 with missing certificate."""
        creds = {'client_cert': '/opt/rucio/etc/web/notthere.crt'}
        BaseClient(account='root', ca_cert=self.cacert, auth_type='x509', creds=creds)

    @raises(ClientProtocolNotSupported)
    def testClientProtocolNotSupported(self):
        """ CLIENTS (BASECLIENT): try to pass an host with a not supported protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        BaseClient(rucio_host='localhost', auth_host='junk://localhost', account='root', auth_type='userpass', creds=creds)


class TestRucioClients():

    def setup(self):
        self.cacert = config_get('test', 'cacert')
        self.marker = '$> '

    def test_ping(self):
        """ PING (CLIENT): Ping Rucio """
        creds = {'username': 'ddmlab', 'password': 'secret'}

        c = Client(account='root', ca_cert=self.cacert, auth_type='userpass', creds=creds)

        print c.ping()



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Thomas Beermann, <thomas.beermann@cern.ch>, 2014-2016
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2015
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2015

"""
This daemon consumes tracer messages from ActiveMQ and updates the atime for replicas.
"""

from datetime import datetime
from dns import resolver
import logging
from os import getpid
from socket import gethostname
from ssl import PROTOCOL_TLSv1
from sys import stdout
from threading import Event, Thread, current_thread
from time import sleep, time
from traceback import format_exc
from Queue import Queue

from json import loads as jloads, dumps as jdumps
from stomp import Connection

from rucio.common.config import config_get, config_get_bool, config_get_int
from rucio.core.monitor import record_counter, record_timer
from rucio.core.did import touch_dids, list_parent_dids
from rucio.core.heartbeat import live, die, sanity_check
from rucio.core.lock import touch_dataset_locks
from rucio.core.replica import touch_replica, touch_collection_replicas
from rucio.db.sqla.constants import DIDType

logging.getLogger("stomp").setLevel(logging.CRITICAL)

logging.basicConfig(stream=stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')

graceful_stop = Event()


class AMQConsumer(object):
    def __init__(self, broker, conn, queue, chunksize, subscription_id, excluded_usrdns, dataset_queue):
        self.__broker = broker
        self.__conn = conn
        self.__queue = queue
        self.__reports = []
        self.__ids = []
        self.__chunksize = chunksize
        self.__subscription_id = subscription_id
        # excluded states empty for the moment, maybe that should be recosidered in the future
        self.__excluded_states = set([])
        # exclude specific usrdns like GangaRBT
        self.__excluded_usrdns = excluded_usrdns
        self.__dataset_queue = dataset_queue

    def on_error(self, headers, message):
        record_counter('daemons.tracer.kronos.error')
        logging.error('[%s] %s' % (self.__broker, message))

    def on_message(self, headers, message):
        record_counter('daemons.tracer.kronos.reports')

        appversion = 'dq2'
        id = headers['message-id']
        if 'appversion' in headers:
            appversion = headers['appversion']

        if 'resubmitted' in headers:
            record_counter('daemons.tracer.kronos.received_resubmitted')
            logging.warning('(kronos_file) got a resubmitted report')

        try:
            if appversion == 'dq2':
                self.__conn.ack(id, self.__subscription_id)
                return
            else:
                report = jloads(message)
        except:
            # message is corrupt, not much to do here
            # send count to graphite, send ack to broker and return
            record_counter('daemons.tracer.kronos.json_error')
            logging.error('(kronos_file) json error')
            self.__conn.ack(id, self.__subscription_id)
            return

        self.__ids.append(id)
        self.__reports.append(report)

        try:
            logging.debug('(kronos_file) message received: %s %s %s' % (str(report['eventType']), report['filename'], report['remoteSite']))
        except:
            pass

        if len(self.__ids) >= self.__chunksize:
            self.__update_atime()
            for id in self.__ids:
                self.__conn.ack(id, self.__subscription_id)

            self.__reports = []
            self.__ids = []

    def __update_atime(self):
        """
        Bulk update atime.
        """
        replicas = []
        rses = []
        for report in self.__reports:
            try:
                # check if scope in report. if not skip this one.
                if 'scope' not in report:
                    record_counter('daemons.tracer.kronos.missing_scope')
                    continue
                else:
                    record_counter('daemons.tracer.kronos.with_scope')

                # for the moment only report with eventType get* are handled.
                if not report['eventType'].startswith('get') and not report['eventType'] == 'download':
                    continue
                record_counter('daemons.tracer.kronos.total_get')
                if report['eventType'] == 'get':
                    record_counter('daemons.tracer.kronos.dq2clients')
                elif report['eventType'] == 'get_sm':
                    if report['eventVersion'] == 'aCT':
                        record_counter('daemons.tracer.kronos.panda_production_act')
                    else:
                        record_counter('daemons.tracer.kronos.panda_production')
                elif report['eventType'] == 'get_sm_a':
                    if report['eventVersion'] == 'aCT':
                        record_counter('daemons.tracer.kronos.panda_analysis_act')
                    else:
                        record_counter('daemons.tracer.kronos.panda_analysis')
                elif report['eventType'] == 'download':
                    record_counter('daemons.tracer.kronos.rucio_download')
                else:
                    record_counter('daemons.tracer.kronos.other_get')

                # check if the report has the right state.
                if report['eventVersion'] != 'aCT':
                    if report['clientState'] in self.__excluded_states:
                        continue

                if report['eventType'] == 'download':
                    report['usrdn'] = report['account']

                if report['usrdn'] in self.__excluded_usrdns:
                    continue
                if 'remoteSite' not in report:
                    continue
                if not report['remoteSite']:
                    continue

                rses = report['remoteSite'].strip().split(',')
                for rse in rses:
                    replicas.append({'name': report['filename'], 'scope': report['scope'], 'rse': rse, 'accessed_at': datetime.utcfromtimestamp(report['traceTimeentryUnix']), 'traceTimeentryUnix': report['traceTimeentryUnix']})
            except (KeyError, AttributeError):
                logging.error(format_exc())
                record_counter('daemons.tracer.kronos.report_error')
                continue

            for did in list_parent_dids(report['scope'], report['filename']):
                if did['type'] != DIDType.DATASET:
                    continue
                # do not update _dis datasets
                if did['scope'] == 'panda' and '_dis' in did['name']:
                    continue
                for rse in rses:
                    self.__dataset_queue.put({'scope': did['scope'], 'name': did['name'], 'did_type': did['type'], 'rse': rse, 'accessed_at': datetime.utcfromtimestamp(report['traceTimeentryUnix'])})

        logging.debug(replicas)

        try:
            ts = time()
            for replica in replicas:
                # if touch replica hits a locked row put the trace back into queue for later retry
                if not touch_replica(replica):
                    resubmit = {'filename': replica['name'], 'scope': replica['scope'], 'remoteSite': replica['rse'], 'traceTimeentryUnix': replica['traceTimeentryUnix'], 'eventType': 'get', 'usrdn': 'someuser', 'clientState': 'DONE'}
                    self.__conn.send(body=jdumps(resubmit), destination=self.__queue, headers={'appversion': 'rucio', 'resubmitted': '1'})
                    record_counter('daemons.tracer.kronos.sent_resubmitted')
                    logging.warning('(kronos_file) hit locked row, resubmitted to queue')
            record_timer('daemons.tracer.kronos.update_atime', (time() - ts) * 1000)
        except:
            logging.error(format_exc())
            record_counter('daemons.tracer.kronos.update_error')

        logging.info('(kronos_file) updated %d replicas' % len(replicas))


def kronos_file(once=False, thread=0, brokers_resolved=None, dataset_queue=None):
    """
    Main loop to consume tracer reports.
    """

    logging.info('tracer consumer starting')

    hostname = gethostname()
    pid = getpid()
    thread = current_thread()

    chunksize = config_get_int('tracer-kronos', 'chunksize')
    prefetch_size = config_get_int('tracer-kronos', 'prefetch_size')
    subscription_id = config_get('tracer-kronos', 'subscription_id')

    use_ssl = True
    try:
        use_ssl = config_get_bool('tracer-kronos', 'use_ssl')
    except:
        pass

    if not use_ssl:
        username = config_get('tracer-kronos', 'username')
        password = config_get('tracer-kronos', 'password')

    excluded_usrdns = set(config_get('tracer-kronos', 'excluded_usrdns').split(','))

    conns = []
    for broker in brokers_resolved:
        if not use_ssl:
            conns.append(Connection(host_and_ports=[(broker, config_get_int('tracer-kronos', 'port'))],
                                    use_ssl=False,
                                    reconnect_attempts_max=config_get_int('tracer-kronos', 'reconnect_attempts')))
        else:
            conns.append(Connection(host_and_ports=[(broker, config_get_int('tracer-kronos', 'port'))],
                                    use_ssl=True,
                                    ssl_key_file=config_get('tracer-kronos', 'ssl_key_file'),
                                    ssl_cert_file=config_get('tracer-kronos', 'ssl_cert_file'),
                                    ssl_version=PROTOCOL_TLSv1,
                                    reconnect_attempts_max=config_get_int('tracer-kronos', 'reconnect_attempts')))

    logging.info('(kronos_file) tracer consumer started')

    sanity_check(executable='kronos-file', hostname=hostname)
    while not graceful_stop.is_set():
        live(executable='kronos-file', hostname=hostname, pid=pid, thread=thread)
        for conn in conns:
            if not conn.is_connected():
                logging.info('(kronos_file) connecting to %s' % conn.transport._Transport__host_and_ports[0][0])
                record_counter('daemons.tracer.kronos.reconnect.%s' % conn.transport._Transport__host_and_ports[0][0].split('.')[0])
                conn.set_listener('rucio-tracer-kronos', AMQConsumer(broker=conn.transport._Transport__host_and_ports[0],
                                                                     conn=conn,
                                                                     queue=config_get('tracer-kronos', 'queue'),
                                                                     chunksize=chunksize,
                                                                     subscription_id=subscription_id,
                                                                     excluded_usrdns=excluded_usrdns,
                                                                     dataset_queue=dataset_queue))
                conn.start()
                if not use_ssl:
                    conn.connect(username, password)
                else:
                    conn.connect()
                conn.subscribe(destination=config_get('tracer-kronos', 'queue'), ack='client-individual', id=subscription_id, headers={'activemq.prefetchSize': prefetch_size})
        sleep(1)

    logging.info('(kronos_file) graceful stop requested')

    for conn in conns:
        try:
            conn.disconnect()
        except:
            pass

    die(executable='rucio-file', hostname=hostname, pid=pid, thread=thread)
    logging.info('(kronos_file) graceful stop done')


def kronos_dataset(once=False, thread=0, dataset_queue=None):
    logging.info('(kronos_dataset) starting')

    hostname = gethostname()
    pid = getpid()
    thread = current_thread()

    dataset_wait = config_get_int('tracer-kronos', 'dataset_wait')
    start = datetime.now()
    while not graceful_stop.is_set():
        live(executable='kronos-dataset', hostname=hostname, pid=pid, thread=thread)
        if (datetime.now() - start).seconds > dataset_wait:
            __update_datasets(dataset_queue)
            start = datetime.now()
        sleep(10)
    # once again for the backlog
    die(executable='rucio-dataset', hostname=hostname, pid=pid, thread=thread)
    logging.info('(kronos_dataset) cleaning dataset backlog before shutdown...')
    __update_datasets(dataset_queue)


def __update_datasets(dataset_queue):
    len_ds = dataset_queue.qsize()
    datasets = {}
    dslocks = {}
    now = time()
    for i in xrange(0, len_ds):
        dataset = dataset_queue.get()
        did = dataset['scope'] + ":" + dataset['name']
        rse = dataset['rse']
        if did not in datasets:
            datasets[did] = dataset['accessed_at']
        else:
            datasets[did] = max(datasets[did], dataset['accessed_at'])

        if rse is None:
            continue
        if did not in dslocks:
            dslocks[did] = {}
        if rse not in dslocks[did]:
            dslocks[did][rse] = dataset['accessed_at']
        else:
            dslocks[did][rse] = max(dataset['accessed_at'], dslocks[did][rse])
    logging.debug('(kronos_dataset) fetched %d datasets from queue (%ds)' % (len_ds, time() - now))

    total, failed, start = 0, 0, time()
    for did, accessed_at in datasets.items():
        scope, name = did.split(':')
        update_did = {'scope': scope, 'name': name, 'type': DIDType.DATASET, 'accessed_at': accessed_at}
        # if update fails, put back in queue and retry next time
        if not touch_dids((update_did,)):
            update_did['rse'] = None
            dataset_queue.put(update_did)
            failed += 1
        total += 1
    logging.debug('(kronos_dataset) did update for %d datasets, %d failed (%ds)' % (total, failed, time() - start))

    total, failed, start = 0, 0, time()
    for did, rses in dslocks.items():
        scope, name = did.split(':')
        for rse, accessed_at in rses.items():
            update_dslock = {'scope': scope, 'name': name, 'rse': rse, 'accessed_at': accessed_at}
            # if update fails, put back in queue and retry next time
            if not touch_dataset_locks((update_dslock,)):
                dataset_queue.put(update_dslock)
                failed += 1
            total += 1
    logging.debug('(kronos_dataset) did update for %d locks, %d failed (%ds)' % (total, failed, time() - start))

    total, failed, start = 0, 0, time()
    for did, rses in dslocks.items():
        scope, name = did.split(':')
        for rse, accessed_at in rses.items():
            update_dslock = {'scope': scope, 'name': name, 'rse': rse, 'accessed_at': accessed_at}
            # if update fails, put back in queue and retry next time
            if not touch_collection_replicas((update_dslock,)):
                dataset_queue.put(update_dslock)
                failed += 1
            total += 1
    logging.debug('(kronos_dataset) did update for %d collection replicas, %d failed (%ds)' % (total, failed, time() - start))


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """
    graceful_stop.set()


def run(once=False, threads=1):
    """
    Starts up the consumer threads
    """
    logging.info('resolving brokers')

    brokers_alias = []
    brokers_resolved = []
    try:
        brokers_alias = [b.strip() for b in config_get('tracer-kronos', 'brokers').split(',')]
    except:
        raise Exception('Could not load brokers from configuration')

    logging.info('resolving broker dns alias: %s' % brokers_alias)

    brokers_resolved = []
    for broker in brokers_alias:
        brokers_resolved.append([str(tmp_broker) for tmp_broker in resolver.query(broker, 'A')])
    brokers_resolved = [item for sublist in brokers_resolved for item in sublist]

    logging.debug('brokers resolved to %s', brokers_resolved)

    dataset_queue = Queue()
    logging.info('starting tracer consumer threads')

    thread_list = []
    for i in xrange(0, threads):
        thread_list.append(Thread(target=kronos_file, kwargs={'thread': i,
                                                              'brokers_resolved': brokers_resolved,
                                                              'dataset_queue': dataset_queue}))
        thread_list.append(Thread(target=kronos_dataset, kwargs={'thread': i,
                                                                 'dataset_queue': dataset_queue}))

    [t.start() for t in thread_list]

    logging.info('waiting for interrupts')

    while len(thread_list) > 0:
        thread_list = [t.join(timeout=3) for t in thread_list if t and t.isAlive()]



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2011-2013
# - Angelos Molfetas, <angelos.molfetas@cern.ch>, 2011
# - Thomas Beermann, <thomas.beermann@cern.ch>, 2012
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2012
# - Martin Barisits, <martin.barisits@cern.ch>, 2014
# - Joaquin Bogado, <joaquin.bogado@cern.ch>, 2015
# - Cedric Serfon, <cedric.serfon@cern.ch>, 2015


import rucio.api.permission
import rucio.common.exception
import rucio.core.identity

from rucio.core import account as account_core
from rucio.common.schema import validate_schema
from rucio.db.sqla.constants import AccountType


def add_account(account, type, email, issuer):
    """
    Creates an account with the provided account name, contact information, etc.

    :param account: The account name.
    :param type: The account type
    :param email: The Email address associated with the account.

    :param issuer: The issuer account_core.

    """

    validate_schema(name='account', obj=account)

    kwargs = {'account': account, 'type': type}
    if not rucio.api.permission.has_permission(issuer=issuer, action='add_account', kwargs=kwargs):
        raise rucio.common.exception.AccessDenied('Account %s can not add account' % (issuer))

    account_core.add_account(account, AccountType.from_sym(type), email)


def del_account(account, issuer):
    """
    Disables an account with the provided account name.

    :param account: The account name.
    :param issuer: The issuer account_core.

    """
    kwargs = {'account': account}
    if not rucio.api.permission.has_permission(issuer=issuer, action='del_account', kwargs=kwargs):
        raise rucio.common.exception.AccessDenied('Account %s can not delete account' % (issuer))

    account_core.del_account(account)


def get_account_info(account):
    """
    Returns the info like the statistics information associated to an account_core.

    :param account: The account name.
    :returns: A list with all account information.
    """
    return account_core.get_account(account)


def get_account_status(account):
    """
    Returns the state of the account_core.

    :param account: Name of the account_core.
    """
    return account_core.get_account_status(account)


def set_account_status(account, status, issuer='root'):
    """ Set the status of an account_core.

    :param account: Name of the account_core.
    :param status: The status for the account_core.
    """
    validate_schema(name='account', obj=account)
    kwargs = {}
    if not rucio.api.permission.has_permission(issuer=issuer, action='set_account_status', kwargs=kwargs):
        raise rucio.common.exception.AccessDenied('Account %s can not change the status of the account' % (issuer))
    return account_core.set_account_status(account, status)


def list_accounts(filter={}):
    """
    Lists all the Rucio account names.

    REST API: http://<host>:<port>/rucio/accounts

    :param filter: Dictionary of attributes by which the input data should be filtered

    :returns: List of all accounts.
    """
    return account_core.list_accounts(filter=filter)


def account_exists(account):
    """
    Checks to see if account exists. This procedure does not check it's status.

    :param account: Name of the account_core.
    :returns: True if found, otherwise false.
    """
    return account_core.account_exists(account)


def list_identities(account):
    """
    List all identities on an account_core.

    :param account: The account name.
    """
    return account_core.list_identities(account)


def list_account_attributes(account):
    """
    Returns all the attributes for the given account.

    :param account: The account name.
    """
    return account_core.list_account_attributes(account)


def add_account_attribute(key, value, account, issuer):
    """
    Add an attribute to an account.

    :param key: attribute key.
    :param value: attribute value.
    :param account: The account name.
    :param issuer: The issuer account_core.
    """
    validate_schema(name='account_attribute', obj=key)
    validate_schema(name='account_attribute', obj=value)

    kwargs = {'account': account, 'key': key, 'value': value}
    if not rucio.api.permission.has_permission(issuer=issuer, action='add_attribute', kwargs=kwargs):
        raise rucio.common.exception.AccessDenied('Account %s can not add attributes' % (issuer))
    account_core.add_account_attribute(account, key, value)


def del_account_attribute(key, account, issuer):
    """
    Delete an attribute to an account.

    :param key: attribute key.
    :param account: The account name.
    :param issuer: The issuer account_core.
    """
    kwargs = {'account': account, 'key': key}
    if not rucio.api.permission.has_permission(issuer=issuer, action='del_attribute', kwargs=kwargs):
        raise rucio.common.exception.AccessDenied('Account %s can not delete attribute' % (issuer))
    account_core.del_account_attribute(account, key)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2014


import logging
import random

from nose.tools import assert_equal

from rucio.api import rse, did
from rucio.core import replica
from rucio.common import exception
from rucio.daemons.cache.consumer import DID_NOT_FOUND, META_MISMATCH
from rucio.daemons.cache.consumer import cache_add_replicas, cache_delete_replicas, RSE_Volatile


class TestRucioCacheFunc():

    def setup(self):
        """RucioCache (Func): Find necessary rse and dids """
        self.id = int(random.random() * 10000)
        self.rse_exist_volatile = 'RUCIO_CACHE_VOLATILE' + str(self.id)
        try:
            rse.add_rse(self.rse_exist_volatile, 'root', deterministic=True, volatile=True)
        except exception.Duplicate:
            logging.warning("rse RUCIO_CACHE_VOLATILE already there")

        self.rse_exist_novolatile = 'RUCIO_CACHE_NOVOLATILE' + str(self.id)
        try:
            rse.add_rse(self.rse_exist_novolatile, 'root', deterministic=True, volatile=False)
        except exception.Duplicate:
            logging.warning("rse RUCIO_CACHE_NOVOLATILE already there")

        self.rse_noExist = 'RUCIO_CACHE_NOEXIST' + str(self.id)
        dids = did.list_dids(scope='mock', filters={}, type='file')
        i = 0
        self.files_exist = []
        self.files_exist_wrong_meta = []
        self.file_replica_on_novolatile = []
        for _did in dids:
            if i < 2:
                i += 1
                meta = did.get_metadata(scope='mock', name=_did[0])
                self.files_exist.append({'scope': meta['scope'], 'name': meta['name'], 'bytes': meta['bytes'], "adler32": meta["adler32"]})
                self.files_exist_wrong_meta.append({'scope': meta['scope'], 'name': meta['name'], 'bytes': 12345678, "adler32": '12345678'})
            elif i < 3:
                meta = did.get_metadata(scope='mock', name=_did[0])
                file = {'scope': meta['scope'], 'name': meta['name'], 'bytes': meta['bytes'], "adler32": meta["adler32"]}
                self.file_replica_on_novolatile.append(file)
                replica.add_replicas(self.rse_exist_novolatile, [file], account='root')

        logging.debug("File Exists: %s " % self.files_exist)
        logging.debug("File Exists with wrong metadata: %s " % self.files_exist_wrong_meta)
        logging.debug("File Exists on volatie rses: " % self.file_replica_on_novolatile)

        self.files_noExist = [{'scope': 'mock', 'name': 'file_notexist', "bytes": 1, "adler32": "0cc737eb"}]
        logging.debug("File not Exists: %s " % self.files_noExist)
        self.account = 'root'
        self.lifetime = 2

    def teardown(self):
        """RucioCache (Func): Clean necessary rse and dids """
        try:
            replica.delete_replicas(self.rse_exist_novolatile, self.file_replica_on_novolatile)
            rse.del_rse(self.rse_exist_volatile, 'root')
            rse.del_rse(self.rse_exist_novolatile, 'root')
        except Exception, e:
            print e

    def test_cache_add_delete_replicas(self):
        """ RucioCache(Func): Test rucio cache add and delete replicas(success) """
        ret = cache_add_replicas(self.rse_exist_volatile, self.files_exist, self.account, self.lifetime)
        assert_equal(ret, 0)

        for file in self.files_exist:
            reps = replica.get_replica(self.rse_exist_volatile, file['scope'], file['name'])
            assert_equal(len(reps) > 0, True)

        ret = cache_delete_replicas(self.rse_exist_volatile, self.files_exist, self.account)
        assert_equal(ret, 0)

        for file in self.files_exist:
            try:
                reps = replica.get_replica(self.rse_exist_volatile, file['scope'], file['name'])
                assert False
            except Exception, e:
                if "No row was found" in str(e):
                    assert True
                else:
                    assert False

    def test_cache_add_replicas_no_rse(self):
        """ RucioCache(Func): Test rucio cache add replicas to not existed rse(failed) """
        try:
            cache_add_replicas(self.rse_noExist, self.files_exist, self.account, self.lifetime)
            assert False
        except exception.RSENotFound:
            assert True

    def test_cache_delete_replicas_no_rse(self):
        """ RucioCache(Func): Test rucio cache delete replicas on not existed rse(failed) """
        try:
            cache_add_replicas(self.rse_noExist, self.files_exist, self.account, self.lifetime)
            assert False
        except exception.RSENotFound:
            assert True

    def test_cache_add_replicas_file_meta_wrong(self):
        """ RucioCache(Func): Test rucio cache add replica with wrong meta data(failed) """
        ret = cache_add_replicas(self.rse_exist_volatile, self.files_exist_wrong_meta, self.account, self.lifetime)
        assert_equal(ret, META_MISMATCH)

        for file in self.files_exist_wrong_meta:
            try:
                replica.get_replica(self.rse_exist_volatile, file['scope'], file['name'])
                assert False
            except Exception, e:
                if "No row was found" in str(e):
                    assert True
                else:
                    assert False

    def test_cache_add_replicas_file_not_exist(self):
        """ RucioCache(Func): Test rucio cache add replica with not existed file(failed) """
        ret = cache_add_replicas(self.rse_exist_volatile, self.files_noExist, self.account, self.lifetime)
        assert_equal(ret, DID_NOT_FOUND)

        for file in self.files_noExist:
            try:
                replica.get_replica(self.rse_exist_volatile, file['scope'], file['name'])
                assert False
            except Exception, e:
                if "No row was found" in str(e):
                    assert True
                else:
                    assert False

    def test_rse_volatie_class(self):
        """ RucioCache(Func): Test rucio cache RSE_Volatile class(success) """
        rse_volatile = RSE_Volatile()
        ret = rse_volatile.get_volatile(self.rse_exist_volatile)
        assert_equal(ret, True)

        ret = rse_volatile.get_volatile(self.rse_exist_novolatile)
        assert_equal(ret, False)



'''
  Copyright European Organization for Nuclear Research (CERN)

  Licensed under the Apache License, Version 2.0 (the "License");
  You may not use this file except in compliance with the License.
  You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0

  Authors:
  - Vincent Garonne, <vincent.garonne@cern.ch>, 2013-2016
'''

from rucio.common.utils import generate_uuid
from rucio.core import rse as rse_core
from rucio.core import replica as replica_core
from rucio.daemons.reaper.reaper import reaper


class TestReaper:
    """ TestReaper Class."""

    def test_reaper(self):
        """ REAPER (DAEMON): Test the reaper daemon."""
        nb_files = 30
        file_size = 2147483648L  # 2G
        for i in xrange(nb_files):
            replica_core.add_replica(rse='MOCK', scope='data13_hip', name='lfn' + generate_uuid(), bytes=file_size, account='root', adler32=None, md5=None)

        rse_core.set_rse_usage(rse='MOCK', source='srm', used=nb_files * file_size, free=800L)
        rse_core.set_rse_limits(rse='MOCK', name='MinFreeSpace', value=10737418240L)
        rse_core.set_rse_limits(rse='MOCK', name='MaxBeingDeletedFiles', value=10)

        rses = [rse_core.get_rse('MOCK'), ]
        reaper(once=True, rses=rses)
        reaper(once=True, rses=rses)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#              http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Ralph Vigne, <ralph.vigne@cern.ch>, 2013


import json
import time
import threading
import traceback

from gearman.client import GearmanClient
from rucio.common.utils import generate_uuid as uuid


class UCEmulator(object):
    """
        This class acts as base class for all use cases included in the performance emulation.
        Every use case module included in rucio.tests.emulation.usecase package and listed
        in the __init__.py under __all__ must be derived from this class.

        One the class is instantiated, it executes the defined use cases (i.e. all methods
        implementing the UseCase decorator) in a frequency according to the current time frame
        defined in the related time series.
    """
    # __ucs = {}

    def __init__(self, cfg=None, carbon_server=None, worker_mode=False):
        """
            Initializes the use case emulation. It sets up the multiplier of the workload defined in the
            time series file and the operation mode for the emulation (both defined in etc/emulation.cfg).

            The name of time series file should be the same as the module defining the use cases and the files
            must be stored either in /opt/rucio/lib/rucio/tests/emulation/timeseries/{modulename}.json or in
            $RUCIO_HOME/lib/rucio/tests/emulation/timeseries/{modulename}.json.

            The operation mode supports 'verbose' and 'threaded'. In 'verbose' mode, only one thread per module is used
            and a lot of debugging information is written to the console. This mode is intended to be used during the development
            of a new use case module. In mode 'threaded', each API call is executed in separate thread and less output is provided.

            :param timeseries_file: the name of the file where the time series for the defined use cases are found.
            :param cfg: the context defined in etc/emulation.cfg
        """
        self.__carbon_server = carbon_server
        if worker_mode:
            return

        if cfg is None:
            print 'Module %s has not configuration' % self.__module__.split('.')[-1]
            return
        self.__intervals = {}
        self.__call_methods = {}
        self.__open_requests = {}

        # Create Context for UseCases and initilaize it with values provided in the cfg - file
        self.__ctx = Context()
        for m in cfg['context']:
            setattr(self.__ctx, m, cfg['context'][m])

        if 'setup' in dir(self):  # Calls setup-method of child class to support the implementation of correlated use cases
            self.setup(self.__ctx)

        for uc in cfg:
            if uc == 'context':
                continue
            else:
                self.__intervals[uc] = 1.0 / cfg[uc]
            try:
                self.__call_methods[uc] = {'main': getattr(self, uc), 'input': None, 'output': None}
                try:
                    self.__call_methods[uc]['input'] = getattr(self, ''.join([uc, '_input']))
                except AttributeError:
                    pass
                try:
                    self.__call_methods[uc]['output'] = getattr(self, ''.join([uc, '_output']))
                except AttributeError:
                    pass
            except Exception, e:
                print e
                print traceback.format_exc()

    def update_ucs(self, ucs):
        for uc in self.__intervals:
            if uc in ucs:
                self.__intervals[uc] = 1.0 / ucs[uc]
        print '== Assigned Frequencies for %s: %s' % (self.__module__.split('.')[-1], self.__intervals)

    def update_ctx(self, key_chain, value):
        """ If the behaviour of this method must be adapted in order to work with a given usecase it must be overwritten in the according UseCaseDefintion class. """
        print '== Updating context: %s -> %s' % (key_chain, value)
        if len(key_chain) == 1:
            setattr(self.__ctx, key_chain[0], value)
        else:
            attr = getattr(self.__ctx, key_chain[0])
            for key in key_chain[1:-1]:
                attr = attr[key]
            attr[key_chain[-1]] = value
        return self.__ctx

    def get_intervals(self):
        """
            A hash array with the interval between each call for each use case e.g. {'UC1': 0.2, 'UC2': 0.5}

            :returns: interval of use cases
        """
        return self.__intervals

    def __iter__(self):
        """
            Generator for workload. Based on the intervals for each use case, the name
            of the next use case is provided when it is time for it to call. When no
            use case is ready for execution, the methods waits until it is time.

            :returns: the ID of the next use case to execute
        """
        pending_calls = {}
        for uc in self.__intervals:
            pending_calls[uc] = self.__intervals[uc] + time.time()  # Set time for first calls. Doing this in bulk could eventually save a lot of time
        nc = min(pending_calls, key=pending_calls.get)
        while not self.__event.is_set():
            yield nc
            # Find next pending call
            pending_calls[nc] += self.__intervals[nc]
            nc = min(pending_calls, key=pending_calls.get)  # Find next pending call. Doing this in bulk could eventually save a lot of time
            sleep = pending_calls[nc] - time.time()  # Calculate sleep till next call
            if sleep > 0.0003:
                self.__event.wait(sleep)
        self.stop()

    def run(self, cfg_global=None, event=None):
        """
            Starts the actual execution of the use cases.

            :param cfg_global: content of the global section in etc/emulation.cfg
            :param event: the event used for thread synchronization
        """
        if event is None:
            return

        self.__event = event
        # Making this assignement outside of the loop saves computing time by avoiding the if inside the loop
        if cfg_global['operation_mode'] == 'threaded':
            do_it = self.run_threaded
        elif cfg_global['operation_mode'] == 'verbose':
            do_it = self.run_verbose
        elif cfg_global['operation_mode'] == 'gearman':
            self.__gearman_client = GearmanClient(cfg_global['gearman'])
            self.__gearman_server = cfg_global['gearman']
            do_it = self.run_gearman
        else:
            raise Exception("Unknown operation mode set")
        for call in self:
            if call and not self.__event.is_set():
                do_it(call)
        return

    def run_gearman(self, call):
        """ Submits the jobs to a gearman queue.

            :param call: Name of the use case to be submitted.
        """
        uc_data = {'uc_name': call, 'class_name': '.'.join([self.__module__, self.__class__.__name__]), 'input_data': {}}
        uc_data['input_data'] = {}
        if self.__call_methods[call]['input'] is not None:
            uc_data['input_data'] = self.__call_methods[call]['input'](self.__ctx)
            if uc_data['input_data'] is None:  # Allows the input - method to save the excution of the actual usecase e.g. if no data is ready for it
                return
        if self.__call_methods[call]['output'] is not None:
            # Gearman call must wait for the return of the job
            # Submitt job in separate thread and wait for response
            id = uuid()
            t = threading.Thread(target=self.await_gearman_results, kwargs={'data': uc_data, 'uuid': id})
            t.daemon = True
            try:
                t.start()
                self.__open_requests[id] = t
            except Exception, e:
                print '!! ERROR !! run_gearman: %s' % e
        else:
            # Gearman job can just be executed, no waiting necessary
            try:
                self.__gearman_client.submit_job(task='execute_uc', data=json.dumps(uc_data), unique=str(uuid()), background=True, max_retries=10, poll_timeout=300)
            except Exception, e:
                print '!! ERROR !! [%s] Unable to submit job to gearman.' % (time.strftime('%H:%M:%S', time.localtime()))
                print e
                print traceback.format_exc()
                self.__event.set()

    def await_gearman_results(self, data, uuid):
        """
            Submits a job to the gearman queue and waits for the response.

            :param data: use case data
        """
        try:
            client = GearmanClient(self.__gearman_server)
            request = None
            try:
                request = client.submit_job(task='execute_uc', data=json.dumps(data), unique=uuid, background=False, wait_until_complete=True)
                if request.state == 'FAILED' and not request.timed_out:  # Can only happen when import on the worker fails
                    print '!! ERROR !! [%s] Worker failed while executing %s.%s (Note: do Rucio imports work for the workers?)' % (time.strftime('%H:%M:%S', time.localtime()), data['class_name'], data['uc_name'])
                    return
                elif request.timed_out:  # Can only happen when import on the worker fails
                    print '!! ERROR !! [%s] Worker timed out while executing %s.%s' % (time.strftime('%H:%M:%S', time.localtime()), data['class_name'], data['uc_name'])
                    return
                elif (request.result.lower() == 'false') or () or (request.result is False):
                    print '!! ERROR !! [%s] %s failed' % (time.strftime('%H:%M:%S', time.localtime()), data['uc_name'])
                    return
                elif request.result is True:
                    print '!! ERROR !! [%s] %s Hmm. Why? ' % (time.strftime('%H:%M:%S', time.localtime()), data['uc_name'])
                    print request.result
                else:
                    result = json.loads(request.result)[1]
                del self.__open_requests[uuid]
            except Exception, e:
                print '!! ERROR !!: %s' % e
                print traceback.format_exc()
                del self.__open_requests[uuid]
                if request:
                    print 'Request-Object: ', request.result
                return
            self.__call_methods[data['uc_name']]['output'](self.__ctx, result)
        except Exception, e:
            print e
            print traceback.format_exc()

    def run_threaded(self, call):
        """
            Starts use case in a new thread.
            IMPORTANT: It is assumed that on the gearman worker the rucio-client package (including the package rucio.test.emulation) is installed
                        and etc/emulation.cfg is properly set up.

            :param call: Name of use case.
        """
        # TODO: Later consider to create a pool of threads for each 'call' and re-use them to potentially save time?
        t = threading.Thread(target=self.time_uc, kwargs={'fn': self.__call_methods[call]})
        t.daemon = True
        try:
            t.start()
        except Exception, e:
            print '!! ERROR !! threaded: %s' % e

    def run_verbose(self, call):
        """
            Executes use case in current thread and prints log data to console.

            :param call: Name of use case.
        """
        print '%f\t%s' % (time.time(), call)
        try:
            self.time_uc(self.__call_methods[call])
        except Exception, e:
            print e
            print traceback.format_exc()
            self.inc('exceptions.%s' % e.__class__.__name__)

    def stop(self):
        """
            Stops the emulation.
        """
        retry = 0
        pending = self.__open_requests.values()
        print '== [%s] Open requests for gearman workers: %s' % (time.strftime('%H:%M:%S', time.localtime()), len(self.__open_requests))
        while len(pending):
            for t in pending:
                if not t.is_alive():
                    pending.remove(t)
            if len(pending):
                if retry > 12:
                    print '!! ERROR !! [%s] Missed %s gearman results due to timeouts during shutdown' % (time.strftime('%H:%M:%S', time.localtime()), len(pending))
                    break
                print '== [%s] Waiting for %s pending gearman responses (retry: %s)' % (time.strftime('%H:%M:%S', time.localtime()), len(pending), retry)
                retry += 1
                time.sleep(10)
        if 'shutdown' in dir(self):  # Calls setup-method of child class to support the implementation of correlated use cases
            self.shutdown(self.__ctx)
        print '= [%s] Stopped module: %s' % (time.strftime('%H:%M:%S', time.localtime()), '.'.join([self.__module__, self.__class__.__name__]))

    def time_uc(self, fn):
        """
            Automatically logs the execution time of t he provided operation.

            :param function: A reference to the operation.
            :param args: Arguments used to execute the operation as list
            :param kwargs: Arguments used to execute the operation as dict

            :returns: function return
        """
        input_data = None
        if 'input' in fn.keys() and fn['input']:
            input_data = fn['input'](self.__ctx)

        res = None
        start = time.time()
        if input_data is not None:
            res = fn['main'](**input_data)
        else:
            res = fn['main']()
        fin = time.time()
        resp = (fin - start) * 1000  # Converting seconds to milliseconds (needed by pystatsd)
        if 'output' in fn.keys() and fn['output']:
            res = fn['output'](self.__ctx, res[1])
        if self.__carbon_server:
            if (fn['main'].__name__ == '__execute__'):  # a wrapped use case is executed, if res is False, an exception occurred and was already reported by the wrapper
                if res:
                    try:
                        self.__carbon_server.timing(res[0], resp)
                    except Exception, e:
                        print '!! ERROR !! Error reporting to grapthite: %s' % e
                    res = res[1]
            else:
                try:
                    self.__carbon_server.timing(fn.__name__, resp)
                except Exception, e:
                    print '!! ERROR !! Error reporting to grapthite: %s' % e
        return res

    def inc(self, metric, value=1):
        """
            Increments the referred metric by value.
        """
        if self.__carbon_server:
            try:
                self.__carbon_server.update_stats(metric, value)
            except Exception, e:
                print '!! ERROR !! Error reporting to grapthite: %s' % e

    @classmethod
    def UseCase(cls, func):
        """
            Decorator to help identifying all use cases defined within a module.
        """
        # Wrap function for exception handling
        def __execute__(self, *args, **kwargs):
            try:
                res = func(self, *args, **kwargs)
                return [func.__name__, res]
            except Exception, e:
                self.inc('exceptions.%s.%s.%s' % (self.__module__.split('.')[-1], func.__name__, e.__class__.__name__))
                print 'Error in: %s.%s: %s' % (self.__module__.split('.')[-1], func.__name__, e)
                print traceback.format_exc()
                return False
        return __execute__


class Context(object):
    #  Needs to be done like this as the pure pythpn object has no dict to save attributes. Cleaner work-arounds are welcome ;-)
    pass

    def __str__(self):
        return str(self.__dict__)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Martin Barisits, <martin.barisits@cern.ch>, 2015

"""
Judge-Injector is a daemon to asynchronously create replication rules
"""

import logging
import os
import socket
import sys
import threading
import time
import traceback

from copy import deepcopy
from datetime import datetime, timedelta
from re import match
from random import randint

from sqlalchemy.exc import DatabaseError

from rucio.common.config import config_get
from rucio.common.exception import DatabaseException, RuleNotFound, RSEBlacklisted, ReplicationRuleCreationTemporaryFailed
from rucio.core.heartbeat import live, die, sanity_check
from rucio.core.rule import inject_rule, get_injected_rules
from rucio.core.monitor import record_counter

graceful_stop = threading.Event()

logging.basicConfig(stream=sys.stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')


def rule_injector(once=False):
    """
    Main loop to check for asynchronous creation of replication rules
    """

    hostname = socket.gethostname()
    pid = os.getpid()
    current_thread = threading.current_thread()

    paused_rules = {}  # {rule_id: datetime}

    # Make an initial heartbeat so that all judge-inectors have the correct worker number on the next try
    live(executable='rucio-judge-injector', hostname=hostname, pid=pid, thread=current_thread, older_than=2 * 60 * 60)
    graceful_stop.wait(1)

    while not graceful_stop.is_set():
        try:
            # heartbeat
            heartbeat = live(executable='rucio-judge-injector', hostname=hostname, pid=pid, thread=current_thread, older_than=2 * 60 * 60)

            start = time.time()
            rules = get_injected_rules(total_workers=heartbeat['nr_threads'] - 1,
                                       worker_number=heartbeat['assign_thread'],
                                       limit=10)
            logging.debug('rule_injector[%s/%s] index query time %f fetch size is %d' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, time.time() - start, len(rules)))

            # Refresh paused rules
            iter_paused_rules = deepcopy(paused_rules)
            for key in iter_paused_rules:
                if datetime.utcnow() > paused_rules[key]:
                    del paused_rules[key]

            # Remove paused rules from result set
            rules = [rule for rule in rules if rule[0] not in paused_rules]

            if not rules and not once:
                logging.debug('rule_injector[%s/%s] did not get any work' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1))
                graceful_stop.wait(60)
            else:
                for rule in rules:
                    rule_id = rule[0]
                    logging.info('rule_injector[%s/%s]: Injecting rule %s' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, rule_id))
                    if graceful_stop.is_set():
                        break
                    try:
                        start = time.time()
                        inject_rule(rule_id=rule_id)
                        logging.debug('rule_injector[%s/%s]: injection of %s took %f' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, rule_id, time.time() - start))
                    except (DatabaseException, DatabaseError), e:
                        if match('.*ORA-00054.*', str(e.args[0])):
                            paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))
                            record_counter('rule.judge.exceptions.LocksDetected')
                            logging.warning('rule_injector[%s/%s]: Locks detected for %s' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, rule_id))
                        elif match('.*QueuePool.*', str(e.args[0])):
                            logging.warning(traceback.format_exc())
                            record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
                        elif match('.*ORA-03135.*', str(e.args[0])):
                            logging.warning(traceback.format_exc())
                            record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
                        else:
                            logging.error(traceback.format_exc())
                            record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
                    except RSEBlacklisted, e:
                        paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))
                        logging.warning('rule_injector[%s/%s]: RSEBlacklisted for rule %s' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, rule_id))
                        record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
                    except ReplicationRuleCreationTemporaryFailed, e:
                        paused_rules[rule_id] = datetime.utcnow() + timedelta(seconds=randint(60, 600))
                        logging.warning('rule_injector[%s/%s]: ReplicationRuleCreationTemporaryFailed for rule %s' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, rule_id))
                        record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
                    except RuleNotFound, e:
                        pass
        except (DatabaseException, DatabaseError), e:
            if match('.*QueuePool.*', str(e.args[0])):
                logging.warning(traceback.format_exc())
                record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
            elif match('.*ORA-03135.*', str(e.args[0])):
                logging.warning(traceback.format_exc())
                record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
            else:
                logging.critical(traceback.format_exc())
                record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
        except Exception, e:
            logging.critical(traceback.format_exc())
            record_counter('rule.judge.exceptions.%s' % e.__class__.__name__)
        if once:
            break

    die(executable='rucio-judge-injector', hostname=hostname, pid=pid, thread=current_thread)


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(once=False, threads=1):
    """
    Starts up the Judge-Injector threads.
    """

    hostname = socket.gethostname()
    sanity_check(executable='rucio-judge-injector', hostname=hostname)

    if once:
        rule_injector(once)
    else:
        logging.info('Injector starting %s threads' % str(threads))
        threads = [threading.Thread(target=rule_injector, kwargs={'once': once}) for i in xrange(0, threads)]
        [t.start() for t in threads]
        # Interruptible joins require a timeout.
        while threads[0].is_alive():
            [t.join(timeout=3.14) for t in threads]



'''
  Copyright European Organization for Nuclear Research (CERN)

  Licensed under the Apache License, Version 2.0 (the "License");
  You may not use this file except in compliance with the License.
  You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0

  Authors:
  - Mario Lassnig, <mario.lassnig@cern.ch>, 2015
'''

from rucio.common.config import config_get
from rucio.core.message import add_message
from rucio.daemons.hermes import hermes


class TestHermes:
    ''' Test the messaging deamon. '''

    def test_hermes(self):
        ''' HERMES (DAEMON): Test the messaging daemon. '''
        for i in xrange(1, 4):
            add_message('test-type_%i' % i, {'test': i})
            add_message('email', {'to': config_get('messaging-hermes', 'email_test').split(','),
                                  'subject': 'Half-Life %i' % i,
                                  'body': '''
                                  Good morning, and welcome to the Black Mesa Transit System.

                                  This automated train is provided for the security and convenience of
                                  the Black Mesa Research Facility personnel. The time is eight-forty
                                  seven A.M... Current outside temperature is ninety three degrees with
                                  an estimated high of one hundred and five. Before exiting the train,
                                  be sure to check your area for personal belongings.

                                  Thank you, and have a very safe, and productive day.'''})

        hermes.run(once=True, send_email=False)



from datetime import datetime
from nose.tools import eq_
from rucio.daemons.auditor import hdfs
from rucio.tests.common import stubbed
import os
import shutil
import string
import tempfile


class FakeHDFSGet:
    def __init__(self, files=[]):
        self.files = files

    def __call__(self, src_path, dst_dir):
        # src_path is ignored, it is assumed a directory is requested
        for content, name in self.files:
            with open(os.path.join(dst_dir, name), 'w') as f:
                f.write(str(content) + '\n')


class TestReplicaFromHDFS:
    def setup(self):
        self.work_dir = tempfile.mkdtemp()

    def teardown(self):
        shutil.rmtree(self.work_dir)

    def test_replica_from_hdfs_download_merges_the_file_parts_in_order(self):
        files = reversed(list(enumerate(string.lowercase[:5])))
        with stubbed(hdfs._hdfs_get, FakeHDFSGet(files)):
            merged_file_path = hdfs.ReplicaFromHDFS.download(
                'FAKE_RSE',
                datetime.now(),
                cache_dir=self.work_dir,
            )

        with open(merged_file_path) as f:
            eq_('01234', f.read().replace('\n', ''))

    def test_replica_from_hdfs_download_reads_files_bigger_than_buffer_size(self):
        files = [
            (string.digits[:5], 'a'),
            (string.digits[5:10], 'b'),
        ]
        with stubbed(hdfs._hdfs_get, FakeHDFSGet(files)):
            merged_file_path = hdfs.ReplicaFromHDFS.download(
                'FAKE_RSE',
                datetime.now(),
                cache_dir=self.work_dir,
                buffer_size=2,
            )

        with open(merged_file_path) as f:
            eq_('0123456789', f.read().replace('\n', ''))



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#              http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Ralph Vigne, <ralph.vigne@cern.ch>, 2013


import fcntl
import json
import multiprocessing
import resource
import operator
import os
import signal
import sys
import time
import threading
import traceback

from gearman.admin_client import GearmanAdminClient
from pystatsd import Client

from rucio.tests.emulation.ucprocess import UCProcess
"""
    Executes all use cases defined in rucio.tests.emulation.usecases according to the time series
    defined in /opt/rucio/lib/rucio/tests/emulation/timeseries.

    Emulation setup is loaded from /opt/rucio/etc/emulation.cfg
"""


def get_open_fds():
    fds = []
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    for fd in range(0, soft):
        try:
            fcntl.fcntl(fd, fcntl.F_GETFD)
        except IOError:
            continue
        fds.append(fd)
    return len(fds)


def observe_gearman_queue(cfg, stop_event):
    acs = {}
    for s in cfg['gearman']:
        acs[s.split(':')[0]] = GearmanAdminClient([s])
    cs = Client(host=cfg['carbon']['CARBON_SERVER'], port=cfg['carbon']['CARBON_PORT'], prefix=cfg['carbon']['USER_SCOPE'])
    count = 0
    pid = os.getpid()
    while stop_event.is_set() is False:
        queue = 0
        try:
            for s in acs:
                try:
                    stat = acs[s].get_status()
                    for task in stat:
                        if task['task'] == 'execute_uc':
                            queue += task['queued']
                            cs.gauge('emulator.counts.gearman.%s' % s, task['queued'])
                except Exception:
                    print '= ERROR: Failed to get stats from %s' % s
                    cs.gauge('emulator.counts.gearman.%s' % s, 0)
            cs.gauge('emulator.counts.files.emulator', get_open_fds())
            if not count % 10:
                print '= (PID: %s [%s]) Gearman-Queue size: %s' % (pid, time.strftime('%H:%M:%S', time.localtime()), queue)
            count += 1
        except Exception:
            print traceback.format_exc()
        try:
            stop_event.wait(10.0)
        except KeyboardInterrupt:
            pass
    print '= Stopping queue observer ... OK'


def main_function():
    update = 500
    num_processes = 4
    stop_event = multiprocessing.Event()
    stop = False
    pid = os.getpid()
    spawned_processes = []
    duration = None

    def signal_handler(signal, frame):
        stop_event.set()
        print '= (PID: %s) [%s] Initiate shutdown of %s processes' % (pid, time.strftime('%H:%M:%S', time.localtime()), len(spawned_processes))
        for n in range(3):
            if not len(spawned_processes):
                print '= (PID: %s) [%s] All processes finished' % (pid, time.strftime('%H:%M:%S', time.localtime()))
                break
            while len(spawned_processes):
                spawned_processes[0].join(10)
                if not spawned_processes[0].is_alive():
                    del spawned_processes[0]
                    if stop_event.is_set():
                        print '= [%s] Process shutdown' % (time.strftime('%H:%M:%S', time.localtime()))
                    else:
                        print '= [%s] Process died' % (time.strftime('%H:%M:%S', time.localtime()))
            print '== [%s] %s processes remaining' % (time.strftime('%H:%M:%S', time.localtime()), len(spawned_processes))
        if len(spawned_processes):
            print '= (PID: %s) [%s] %s processes did not finish properly. Kill them now' % (pid, time.strftime('%H:%M:%S', time.localtime()), len(spawned_processes))
            for p in spawned_processes:
                p.terminate()
                print '= [%s] Process killed' % (time.strftime('%H:%M:%S', time.localtime()))
        if timeout_event.is_set():
            sys.exit(1)
        else:
            sys.exit(0)

    signal.signal(signal.SIGTERM, signal_handler)

    with open('/opt/rucio/etc/emulation.cfg') as f:
        cfg = json.load(f)

    # Check sysargs for modules
    exec_mods = list()

    if '--help' in sys.argv:
        print 'Spported command line arguments:'
        print '\t--exclude uc1 uc2 ... ucN\texclude the provided usecase modules from emulation'
        print '\t--only uc1 uc2 ... ucN\tncludes only the provided usecase modules for emulation'
    elif '--only' in sys.argv:
        for mod in cfg['global']['modules']:
            if mod in sys.argv:
                exec_mods.append(mod)

    elif '--exclude' in sys.argv:
        for mod in cfg['global']['modules']:
            if mod not in sys.argv:
                exec_mods.append(mod)
    else:
        exec_mods = cfg['global']['modules']

    # Load module sepcific configurations
    for mod in exec_mods:
        with open('/opt/rucio/etc/%s.cfg' % mod) as f:
            mcfg = json.load(f)
        cfg.update(mcfg)

    print '=' * 80
    print '=' * 35 + ' SETTINGS ' + '=' * 35
    # Printing configuration for testrun
    for setting in cfg['global']:
        print '= Emulation -> %s:\t %s' % (setting, cfg['global'][setting])
        if setting == 'update_interval':
            update = cfg['global'][setting]
        if setting == 'processes':
            num_processes = cfg['global'][setting]
        if setting == 'duration':
            duration = cfg['global'][setting]

    print '=' * 31 + ' INCLUDED USECASES ' + '=' * 30
    uc_array = dict()

    # Distributing WL over processes
    proc_load = {}
    proc_mod = {}
    for i in range(num_processes):
        proc_load[i] = 0
        proc_mod[i] = []

    for mod in cfg:
        if mod == 'global':
            continue
        uc_array[mod] = 0
        for uc in cfg[mod]:
            if uc == 'context':
                continue
            uc_array[mod] += cfg[mod][uc]
    sorted_uc = sorted(uc_array.iteritems(), key=operator.itemgetter(1))
    while len(sorted_uc):
        smallest = sorted(proc_load.iteritems(), key=operator.itemgetter(1))[0][0]
        uc = sorted_uc.pop()
        proc_load[smallest] += uc[1]
        proc_mod[smallest].append(uc[0])

    # Start Gearman queue observer
    if cfg['global']['operation_mode'] == 'gearman':
        print '=' * 36 + ' GEARMAN ' + '=' * 35
        try:
            t = threading.Thread(target=observe_gearman_queue, args=[cfg['global'], stop_event])
            t.deamon = True
            t.start()
            print '= Setting up gearman queue observer ... OK'
        except Exception, e:
            print '!! Unable to connect to Gearman-Server !!'
            print e
            print traceback.format_exc()

    # Starting all processes
    procs = []
    for proc in proc_mod:
        if len(proc_mod[proc]):
            # ucp = UCProcess(cfg, proc_mod[proc], multiprocessing.Event())
            ucp = UCProcess(cfg, proc_mod[proc], stop_event)
            procs.append(ucp)
    print '=' * 30 + ' STARTING EXECUTION ' + '=' * 30
    if len(procs) > 1:
        for ucp in procs:
            p = multiprocessing.Process(target=ucp.run)
            p.deamon = True
            p.start()
            spawned_processes.append(p)
    else:
        print '= (PID: %s) Only one module found, unsing threads instead of subprocesses' % (os.getpid())
        p = threading.Thread(target=ucp.run)
        p.deamon = True
        p.start()
        spawned_processes.append(p)

    timeout_event = threading.Event()
    if duration:
        print '== Set timeout for emulation: %.1f minutes' % (duration)
        t = threading.Thread(target=waiting_to_stop, kwargs={'duration': duration, 'interval': update, 'stop_event': stop_event, 'timeout': timeout_event})
        t.daemon = True
        t.start()
    else:
        print '== Disable timeout for emulation (infinite run)'
    while t.is_alive():
        t.join(3)
        if stop:
            stop_event.set()


def waiting_to_stop(duration, interval, stop_event, timeout):
    until = time.time() + duration
    while True:
        try:
            with open('emulator.stop'):
                pass
            print '= [%s] Found stop file (emulator.stop) and shutting down. To restart delete this file.' % (time.strftime('%H:%M:%S', time.localtime()))
            break
        except IOError:
            pass
        if until < time.time():
            print '= [%s] Emulation timed out. Going to shudown.' % (time.strftime('%H:%M:%S', time.localtime()))
            timeout.set()
            break
        if stop_event.is_set():
            print '= Received stop signal.'
            break
        stop_event.wait(interval)
        print '= [%s] Time remaining before shuting down emulation: %s minutes' % (time.strftime('%H:%M:%S', time.localtime()), ((until - time.time()) / 60))

    print '= [%s] Shutting down emulation' % (time.strftime('%H:%M:%S', time.localtime()))
    stop_event.set()
    print '= [%s] Proper shutdown done.' % (time.strftime('%H:%M:%S', time.localtime()))


if __name__ == '__main__':
    return_value = main_function()
    print '== Proper shutdown finished'
    sys.exit(return_value)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2014
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2016

import os
import urlparse

import boto
import boto.s3.connection

from boto.s3.key import Key

from rucio.common import exception
from rucio.common.config import get_rse_credentials

from rucio.rse.protocols import protocol


class Default(protocol.RSEProtocol):
    """ Implementing access to RSEs using the S3 protocol."""

    def __init__(self, protocol_attr, rse_settings):
        super(Default, self).__init__(protocol_attr, rse_settings)
        if 'determinism_type' in self.attributes:
            self.attributes['determinism_type'] = 's3'
        self.__conn = None

    def _get_path(self, scope, name):
        """ Transforms the physical file name into the local URI in the referred RSE.
            Suitable for sites implementoing the RUCIO naming convention.

            :param name: filename
            :param scope: scope

            :returns: RSE specific URI of the physical file
        """
        return '%s:%s' % (scope, name)

    def get_bucket_key_name(self, pfn):
        """
            Gets boto key for a pfn

            :param pfn: Physical file name

            :returns: bucket name and key name as string
        """
        try:
            parsed = urlparse.urlparse(pfn)
            hash_path = parsed.path.strip("/")

            pos = hash_path.index("/")
            bucket_name = hash_path[:pos]
            key_name = hash_path[pos + 1:]

            return bucket_name, key_name
        except Exception as e:
            raise exception.RucioException(str(e))

    def get_bucket_key(self, pfn, create=False):
        """
            Gets boto key for a pfn

            :param pfn: Physical file name
            :param create: True if needs to create the key, False if not

            :returns: boto bucket and key object
        """
        try:
            bucket_name, key_name = self.get_bucket_key_name(pfn)

            if create:
                bucket = self.__conn.create_bucket(bucket_name)
                key = Key(bucket, key_name)
            else:
                bucket = self.__conn.get_bucket(bucket_name)
                key = bucket.get_key(key_name)
            return bucket, key
        except boto.exception.S3ResponseError as e:
            if e.status == 404:
                raise exception.SourceNotFound(str(e))
            else:
                raise exception.ServiceUnavailable(e)

    def exists(self, path):
        """
            Checks if the requested file is known by the referred RSE.

            :param path: Physical file name

            :returns: True if the file exists, False if it doesn't

            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        try:
            bucket, key = self.get_bucket_key(path)
            if key:
                return True
            else:
                return False
        except exception.SourceNotFound:
            return False
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def connect(self):
        """
            Establishes the actual connection to the referred RSE.

            :param: credentials needed to establish a connection with the stroage.

            :raises RSEAccessDenied: if no connection could be established.
        """
        try:
            scheme, prefix = self.attributes.get('scheme'), self.attributes.get('prefix')
            netloc, port = self.attributes['hostname'], self.attributes.get('port', 80)
            service_url = '%(scheme)s://%(netloc)s:%(port)s%(prefix)s' % locals()
            credentials = get_rse_credentials()
            self.rse['credentials'] = credentials.get(self.rse['rse'])
            is_secure = self.rse['credentials'].get('is_secure', {}).\
                get(service_url, False)
            self.__conn = boto.connect_s3(host=self.attributes['hostname'],
                                          port=int(port),
                                          aws_access_key_id=self.rse['credentials']['access_key'],
                                          aws_secret_access_key=self.rse['credentials']['secret_key'],
                                          is_secure=is_secure,
                                          calling_format=boto.s3.connection.OrdinaryCallingFormat())
        except Exception as e:
            raise exception.RSEAccessDenied(e)

    def close(self):
        """ Closes the connection to RSE."""
        pass

    def get(self, pfn, dest):
        """
            Provides access to files stored inside connected the RSE.

            :param path: Physical file name of requested file
            :param dest: Name and path of the files when stored at the client

            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occured in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
         """
        try:
            bucket, key = self.get_bucket_key(pfn)
            if key is None:
                raise exception.SourceNotFound('Cannot get the source key from S3')
            key.get_contents_to_filename(dest)
        except IOError as e:
            if e.errno == 2:
                raise exception.DestinationNotAccessible(e)
            else:
                raise exception.ServiceUnavailable(e)
        except exception.SourceNotFound as e:
            raise exception.SourceNotFound(e)
        except Exception as e:
            if os.path.exists(dest):
                os.remove(dest)
            raise exception.ServiceUnavailable(e)

    def put(self, source, target, source_dir=None):
        """
            Allows to store files inside the referred RSE.

            :param source: path to the source file on the client file system
            :param target: path to the destination file on the storage
            :param source_dir: Path where the to be transferred files are stored in the local file system

            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occured in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        full_name = source_dir + '/' + source if source_dir else source
        try:
            bucket, key = self.get_bucket_key(target, create=True)
            if key is None:
                raise exception.DestinationNotAccessible('Cannot get the destionation key from S3')
            key.set_contents_from_filename(full_name)
        except exception.SourceNotFound as e:
            raise exception.SourceNotFound(e)
        except Exception as e:
            if 'No such file' in str(e):
                raise exception.SourceNotFound(e)
            else:
                raise exception.ServiceUnavailable(e)

    def delete(self, pfn):
        """
            Deletes a file from the connected RSE.

            :param path: path to the to be deleted file

            :raises ServiceUnavailable: if some generic error occured in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        try:
            bucket, key = self.get_bucket_key(pfn)
            if key is None:
                raise exception.SourceNotFound('Cannot get the key from S3')
            key.delete()
        except exception.SourceNotFound as e:
            raise exception.SourceNotFound(e)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param path: path to the current file on the storage
            :param new_path: path to the new file on the storage

            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occured in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        try:
            bucket, key = self.get_bucket_key(pfn)
            if key is None:
                raise exception.SourceNotFound('Cannot get the source key from S3')
            bucket_name, key_name = self.get_bucket_key_name(new_pfn)
            key.copy(bucket_name, key_name)
            key.delete()
        except exception.SourceNotFound as e:
            raise exception.SourceNotFound(e)
        except boto.exception.S3ResponseError as e:
            if e.status in [404, 403]:
                raise exception.DestinationNotAccessible(e)
            else:
                raise exception.ServiceUnavailable(e)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def stat(self, pfn):
        """ Determines the file size in bytes  of the provided file.

            :param pfn: The PFN the file.

            :returns: a dict containing the key filesize.
        """
        try:
            bucket, key = self.get_bucket_key(pfn)
            if key is None:
                raise exception.SourceNotFound('Cannot get the key from S3')
            return {'filesize': int(key.size)}
        except exception.SourceNotFound as e:
            raise exception.SourceNotFound(e)
        except Exception as e:
            raise exception.ServiceUnavailable(e)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2013


from rucio.db.sqla.session import get_session


class TestDB():

    def test_db_connection(self):
        """ DB (CORE): Test db connection """
        session = get_session()
        if session.bind.dialect.name == 'oracle':
            session.execute('select 1 from dual')
        else:
            session.execute('select 1')
        session.close()



'''
  Copyright European Organization for Nuclear Research (CERN)

  Licensed under the Apache License, Version 2.0 (the "License");
  You may not use this file except in compliance with the License.
  You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

  Authors:
  - Angelos Molfetas, <angelos.molfetas@cern.ch>, 2012
  - Mario Lassnig, <mario.lassnig@cern.ch>, 2012
  - Vincent Garonne,  <vincent.garonne@cern.ch>, 2011-2016
  - Cedric Serfon, <cedric.serfon@cern.ch>, 2014
'''

import sys

from ConfigParser import NoOptionError
from functools import wraps
from inspect import isgeneratorfunction
# from retrying import retry
from threading import Lock
from os.path import basename

from sqlalchemy import create_engine, event
from sqlalchemy.exc import DatabaseError, DisconnectionError, OperationalError, TimeoutError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

from rucio.common.config import config_get
from rucio.common.exception import RucioException, DatabaseException


BASE = declarative_base()
try:
    DEFAULT_SCHEMA_NAME = config_get('database', 'schema')
    BASE.metadata.schema = DEFAULT_SCHEMA_NAME
except NoOptionError:
    DEFAULT_SCHEMA_NAME = None

_MAKER, _ENGINE, _LOCK = None, None, Lock()


def _fk_pragma_on_connect(dbapi_con, con_record):
    # Hack for previous versions of sqlite3
    try:
        dbapi_con.execute('pragma foreign_keys=ON')
    except AttributeError:
        pass


def mysql_ping_listener(dbapi_conn, connection_rec, connection_proxy):
    """
    Ensures that MySQL connections checked out of the
    pool are alive.

    Borrowed from:
    http://groups.google.com/group/sqlalchemy/msg/a4ce563d802c929f

    :param dbapi_conn: DBAPI connection
    :param connection_rec: connection record
    :param connection_proxy: connection proxy
    """

    try:
        dbapi_conn.cursor().execute('select 1')
    except dbapi_conn.OperationalError, ex:
        if ex.args[0] in (2006, 2013, 2014, 2045, 2055):
            msg = 'Got mysql server has gone away: %s' % ex
            raise DisconnectionError(msg)
        else:
            raise


def my_on_connect(dbapi_con, connection_record):
    """ Adds information to track performance and ressource by module.
        Info are recorded in the V$SESSION and V$SQLAREA views.
    """
    caller = basename(sys.argv[0])
    dbapi_con.clientinfo = caller
    dbapi_con.client_identifier = caller
    dbapi_con.module = caller
    dbapi_con.action = caller


def get_engine(echo=True):
    """ Creates a engine to a specific database.
        :returns: engine
    """
    global _ENGINE
    if not _ENGINE:
        sql_connection = config_get('database', 'default')
        config_params = [('pool_size', int), ('max_overflow', int), ('pool_timeout', int),
                         ('pool_recycle', int), ('echo', int), ('echo_pool', str),
                         ('pool_reset_on_return', str), ('use_threadlocal', int)]
        params = {}
        for param, param_type in config_params:
            try:
                params[param] = param_type(config_get('database', param))
            except NoOptionError:
                pass
        _ENGINE = create_engine(sql_connection, **params)
        if 'mysql' in sql_connection:
            event.listen(_ENGINE, 'checkout', mysql_ping_listener)
        elif 'sqlite' in sql_connection:
            event.listen(_ENGINE, 'connect', _fk_pragma_on_connect)
        elif 'oracle' in sql_connection:
            event.listen(_ENGINE, 'connect', my_on_connect)
    assert _ENGINE
    return _ENGINE


def get_dump_engine(echo=False):
    """ Creates a dump engine to a specific database.
        :returns: engine """

    statements = list()

    def dump(sql, *multiparams, **params):
        statement = str(sql.compile(dialect=engine.dialect))
        if statement in statements:
            return
        statements.append(statement)
        if statement.endswith(')\n\n'):
            if engine.dialect.name == 'oracle':
                print statement.replace(')\n\n', ') PCTFREE 0;\n')
            else:
                print statement.replace(')\n\n', ');\n')
        elif statement.endswith(')'):
            print statement.replace(')', ');\n')
        else:
            print statement
    sql_connection = config_get('database', 'default')

    engine = create_engine(sql_connection, echo=echo, strategy='mock', executor=dump)
    return engine


def get_maker():
    """
        Return a SQLAlchemy sessionmaker.
        May assign __MAKER if not already assigned.
    """
    global _MAKER, _ENGINE
    assert _ENGINE
    if not _MAKER:
        _MAKER = sessionmaker(bind=_ENGINE, autocommit=False, autoflush=True, expire_on_commit=True)
    return _MAKER


def get_session():
    """ Creates a session to a specific database, assumes that schema already in place.
        :returns: session
    """
    global _MAKER, _LOCK
    if not _MAKER:
        _LOCK.acquire()
        try:
            get_engine()
            get_maker()
        finally:
            _LOCK.release()
    assert _MAKER
    session = scoped_session(_MAKER)
    return session


def retry_if_db_connection_error(exception):
    """Return True if error in connecting to db."""
    print exception
    if isinstance(exception, OperationalError):
        conn_err_codes = ('2002', '2003', '2006',  # MySQL
                          'ORA-00028',  # Oracle session has been killed
                          'ORA-01012',  # not logged on
                          'ORA-03113',  # end-of-file on communication channel
                          'ORA-03114',  # not connected to ORACLE
                          'ORA-03135',  # connection lost contact
                          'ORA-25408',)  # can not safely replay call
        for err_code in conn_err_codes:
            if exception.args[0].find(err_code) != -1:
                return True
    return False


def read_session(function):
    '''
    decorator that set the session variable to use inside a function.
    With that decorator it's possible to use the session variable like if a global variable session is declared.

    session is a sqlalchemy session, and you can get one calling get_session().
    This is useful if only SELECTs and the like are being done; anything involving
    INSERTs, UPDATEs etc should use transactional_session.
    '''
#     @retry(retry_on_exception=retry_if_db_connection_error,
#            wait_fixed=0.5,
#            stop_max_attempt_number=2,
#            wrap_exception=False)
    @wraps(function)
    def new_funct(*args, **kwargs):

        if isgeneratorfunction(function):
            raise RucioException('read_session decorator should not be used with generator. Use stream_session instead.')

        if not kwargs.get('session'):
            session = get_session()
            try:
                kwargs['session'] = session
                return function(*args, **kwargs)
            except TimeoutError, error:
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except DatabaseError, error:
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except:
                session.rollback()  # pylint: disable=maybe-no-member
                raise
            finally:
                session.remove()
        try:
            return function(*args, **kwargs)
        except:
            raise
    new_funct.__doc__ = function.__doc__
    return new_funct


def stream_session(function):
    '''
    decorator that set the session variable to use inside a function.
    With that decorator it's possible to use the session variable like if a global variable session is declared.

    session is a sqlalchemy session, and you can get one calling get_session().
    This is useful if only SELECTs and the like are being done; anything involving
    INSERTs, UPDATEs etc should use transactional_session.
    '''
#     @retry(retry_on_exception=retry_if_db_connection_error,
#            wait_fixed=0.5,
#            stop_max_attempt_number=2,
#            wrap_exception=False)
    @wraps(function)
    def new_funct(*args, **kwargs):

        if not isgeneratorfunction(function):
            raise RucioException('stream_session decorator should be used only with generator. Use read_session instead.')

        if not kwargs.get('session'):
            session = get_session()
            try:
                kwargs['session'] = session
                for row in function(*args, **kwargs):
                    yield row
            except TimeoutError, error:
                print error
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except DatabaseError, error:
                print error
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except:
                session.rollback()  # pylint: disable=maybe-no-member
                raise
            finally:
                session.remove()
        else:
            try:
                for row in function(*args, **kwargs):
                    yield row
            except:
                raise
    new_funct.__doc__ = function.__doc__
    return new_funct


def transactional_session(function):
    '''
    decorator that set the session variable to use inside a function.
    With that decorator it's possible to use the session variable like if a global variable session is declared.

    session is a sqlalchemy session, and you can get one calling get_session().
    '''
    @wraps(function)
    def new_funct(*args, **kwargs):
        if not kwargs.get('session'):
            session = get_session()
            try:
                kwargs['session'] = session
                result = function(*args, **kwargs)
                session.commit()  # pylint: disable=maybe-no-member
            except TimeoutError, error:
                print error
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except DatabaseError, error:
                print error
                session.rollback()  # pylint: disable=maybe-no-member
                raise DatabaseException(str(error))
            except:
                session.rollback()  # pylint: disable=maybe-no-member
                raise
            finally:
                session.remove()  # pylint: disable=maybe-no-member
        else:
            result = function(*args, **kwargs)
        return result
    new_funct.__doc__ = function.__doc__
    return new_funct



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2012-2014
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2013-2015
# - Cedric Serfon, <cedric.serfon@cern.ch>, 2014
# - Wen Guan, <wen.guan@cern.ch>, 2014-2016

"""
Conveyor is a daemon to manage file transfers.
"""

import datetime
import json
import logging
import os
import socket
import sys
import threading
import time
import traceback

from collections import defaultdict
from ConfigParser import NoOptionError
from threadpool import ThreadPool, makeRequests

from rucio.common.config import config_get
from rucio.common.utils import chunks
from rucio.core import request, heartbeat
from rucio.core.monitor import record_timer
from rucio.daemons.conveyor import common
from rucio.db.sqla.constants import RequestState, RequestType


logging.basicConfig(stream=sys.stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')

graceful_stop = threading.Event()

# http://bugs.python.org/issue7980
datetime.datetime.strptime('', '')


def poller(once=False,
           process=0, total_processes=1, thread=0, total_threads=1, activities=None, sleep_time=60,
           fts_bulk=100, db_bulk=1000, older_than=60, activity_shares=None):
    """
    Main loop to check the status of a transfer primitive with a transfertool.
    """

    try:
        timeout = config_get('conveyor', 'poll_timeout')
        timeout = float(timeout)
    except NoOptionError:
        timeout = None

    logging.info('poller starting - process (%i/%i) thread (%i/%i) bulk (%i) timeout (%s)' % (process, total_processes,
                                                                                              thread, total_threads,
                                                                                              db_bulk, timeout))

    executable = ' '.join(sys.argv)
    hostname = socket.getfqdn()
    pid = os.getpid()
    hb_thread = threading.current_thread()
    heartbeat.sanity_check(executable=executable, hostname=hostname)
    hb = heartbeat.live(executable, hostname, pid, hb_thread)

    logging.info('poller started - process (%i/%i) thread (%i/%i) bulk (%i)' % (process, total_processes,
                                                                                hb['assign_thread'], hb['nr_threads'],
                                                                                db_bulk))

    activity_next_exe_time = defaultdict(time.time)
    threadPool = ThreadPool(total_threads)
    sleeping = False

    while not graceful_stop.is_set():

        try:
            hb = heartbeat.live(executable, hostname, pid, hb_thread, older_than=3600)
            logging.debug('poller - thread (%i/%i)' % (hb['assign_thread'], hb['nr_threads']))

            if not sleeping:
                sleeping = True

            if activities is None:
                activities = [None]
            for activity in activities:
                if activity_next_exe_time[activity] > time.time():
                    time.sleep(1)
                    continue
                sleeping = False

                ts = time.time()
                logging.debug('%i:%i - start to poll transfers older than %i seconds for activity %s' % (process, hb['assign_thread'], older_than, activity))
                transfs = request.get_next_transfers(request_type=[RequestType.TRANSFER, RequestType.STAGEIN, RequestType.STAGEOUT],
                                                     state=[RequestState.SUBMITTED],
                                                     limit=db_bulk,
                                                     older_than=datetime.datetime.utcnow() - datetime.timedelta(seconds=older_than),
                                                     process=process, total_processes=total_processes,
                                                     thread=hb['assign_thread'], total_threads=hb['nr_threads'],
                                                     activity=activity,
                                                     activity_shares=activity_shares)
                record_timer('daemons.conveyor.poller.000-get_next_transfers', (time.time() - ts) * 1000)

                if transfs:
                    logging.debug('%i:%i - polling %i transfers for activity %s' % (process, hb['assign_thread'], len(transfs), activity))

                xfers_ids = {}
                for transf in transfs:
                    if not transf['external_host'] in xfers_ids:
                        xfers_ids[transf['external_host']] = []
                    xfers_ids[transf['external_host']].append(transf['external_id'])

                for external_host in xfers_ids:
                    for xfers in chunks(xfers_ids[external_host], fts_bulk):
                        # poll transfers
                        # xfer_requests = makeRequests(common.poll_transfers, args_list=[((external_host, xfers, process, thread), {})])
                        xfer_requests = makeRequests(common.poll_transfers, args_list=[((), {'external_host': external_host, 'xfers': xfers, 'process': process, 'thread': hb['assign_thread'], 'timeout': timeout})])
                        [threadPool.putRequest(xfer_req) for xfer_req in xfer_requests]
                threadPool.wait()

                if len(transfs) < db_bulk / 2:
                    logging.info("%i:%i - only %s transfers for activity %s, which is less than half of the bulk %s, will sleep %s seconds" % (process, hb['assign_thread'], len(transfs), activity, db_bulk, sleep_time))
                    if activity_next_exe_time[activity] < time.time():
                        activity_next_exe_time[activity] = time.time() + sleep_time
        except:
            logging.critical("%i:%i - %s" % (process, hb['assign_thread'], traceback.format_exc()))

        if once:
            break

    logging.info('%i:%i - graceful stop requests' % (process, hb['assign_thread']))

    threadPool.dismissWorkers(total_threads, do_join=True)
    heartbeat.die(executable, hostname, pid, hb_thread)

    logging.info('%i:%i - graceful stop done' % (process, hb['assign_thread']))


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(once=False,
        process=0, total_processes=1, total_threads=1, sleep_time=60, activities=None,
        fts_bulk=100, db_bulk=1000, older_than=60, activity_shares=None):
    """
    Starts up the conveyer threads.
    """

    if activity_shares:

        try:
            activity_shares = json.loads(activity_shares)
        except:
            logging.critical('activity share is not a valid JSON dictionary')
            return

        try:
            if round(sum(activity_shares.values()), 2) != 1:
                logging.critical('activity shares do not sum up to 1, got %s - aborting' % round(sum(activity_shares.values()), 2))
                return
        except:
            logging.critical('activity shares are not numbers? - aborting')
            return

        activity_shares.update((share, int(percentage * db_bulk)) for share, percentage in activity_shares.items())
        logging.info('activity shares enabled: %s' % activity_shares)

    if once:
        logging.info('executing one poller iteration only')
        poller(once=once, fts_bulk=fts_bulk, db_bulk=db_bulk, older_than=older_than, activities=activities, activity_shares=activity_shares)

    else:

        logging.info('starting poller threads')

        threads = [threading.Thread(target=poller, kwargs={'process': process,
                                                           'total_processes': total_processes,
                                                           'thread': 0,
                                                           'total_threads': total_threads,
                                                           'older_than': older_than,
                                                           'fts_bulk': fts_bulk,
                                                           'db_bulk': db_bulk,
                                                           'sleep_time': sleep_time,
                                                           'activities': activities,
                                                           'activity_shares': activity_shares})]

        [t.start() for t in threads]

        logging.info('waiting for interrupts')

        # Interruptible joins require a timeout.
        while len(threads) > 0:
            threads = [t.join(timeout=3.14) for t in threads if t and t.isAlive()]



#!/usr/bin/env python
# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2013

from json import dumps
from web import application, data, header, OK

from rucio.tests.mock.fts3 import submit, query, cancel

"""
This mock FTS3 interface implements the following functionality:
 Submit job: POST /jobs
 Query job: GET /<jobid>
 Cancel job: DELETE /<jobid>
 Version:  GET /
"""

urls = (
    '/', 'Version',
    '/jobs', 'Submit',
    '/jobs/(.+)', 'QueryCancel'
)


class Submit:

    def POST(self):
        """
        Create a new transfer job.

        HTTP Success:
            200 OK

        :returns: JSON-encoded transfer job identifier
        """

        header('Content-Type', 'application/json')

        return dumps(submit(data()))


class Version:

    def GET(self):
        """
        Return FTS version information.

        HTTP Success:
            200 OK

        :returns: JSON-encoded version information.
        """

        header('Content-Type', 'application/json')

        return dumps({"api": {"major": 0, "minor": 0, "patch": 0},
                      "delegation": {"major": 0, "minor": 0, "patch": 0},
                      "schema": {"major": 0, "minor": 0, "patch": 0}})


class QueryCancel:

    def GET(self, tid):
        """
        Query transfer job information.

        HTTP Success:
            200 OK

        :param tid: JSON-encoded transfer job identifier.
        :returns: JSON-encoded transfer job information.
        """

        header('Content-Type', 'application/json')

        return dumps(query(tid))

    def DELETE(self, tid):
        """
        Kill a transfer job.

        HTTP Success:
            200 OK

        :param tid: JSON-encoded transfer job identifier.
        """

        header('Content-Type', 'application/octet-stream')
        cancel(tid)
        raise OK()


"""----------------------
   Web service startup
----------------------"""

app = application(urls, globals())
application = app.wsgifunc()



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2015
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2015
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2015


"""
Conveyor finisher is a daemon to update replicas and rules based on requests.
"""

import datetime
import logging
import os
import re
import socket
import sys
import threading
import time
import traceback

from sqlalchemy.exc import DatabaseError

from rucio.common.config import config_get
from rucio.common.utils import chunks
from rucio.common.exception import DatabaseException
from rucio.core import request, heartbeat
from rucio.core.monitor import record_timer, record_counter
from rucio.daemons.conveyor import common
from rucio.db.sqla.constants import RequestState, RequestType


logging.basicConfig(stream=sys.stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')

graceful_stop = threading.Event()


def finisher(once=False, process=0, total_processes=1, thread=0, total_threads=1, sleep_time=60, activities=None, bulk=100, db_bulk=1000):
    """
    Main loop to update the replicas and rules based on finished requests.
    """

    logging.info('finisher starting - process (%i/%i) thread (%i/%i) db_bulk(%i) bulk (%i)' % (process, total_processes,
                                                                                               thread, total_threads,
                                                                                               db_bulk, bulk))
    executable = ' '.join(sys.argv)
    hostname = socket.getfqdn()
    pid = os.getpid()
    hb_thread = threading.current_thread()
    heartbeat.sanity_check(executable=executable, hostname=hostname)
    # Make an initial heartbeat so that all finishers have the correct worker number on the next try
    hb = heartbeat.live(executable, hostname, pid, hb_thread)
    graceful_stop.wait(1)
    sleeping = False
    while not graceful_stop.is_set():

        try:
            hb = heartbeat.live(executable, hostname, pid, hb_thread, older_than=3600)
            logging.debug('finisher - thread (%i/%i)' % (hb['assign_thread'], hb['nr_threads']))

            if activities is None:
                activities = [None]

            if sleeping:
                logging.info('%i:%i - nothing to do. will sleep 60s' % (process, hb['assign_thread']))
                time.sleep(60)

            sleeping = True
            for activity in activities:
                ts = time.time()
                reqs = request.get_next(request_type=[RequestType.TRANSFER, RequestType.STAGEIN, RequestType.STAGEOUT],
                                        state=[RequestState.DONE, RequestState.FAILED, RequestState.LOST, RequestState.SUBMITTING,
                                               RequestState.SUBMISSION_FAILED, RequestState.NO_SOURCES, RequestState.ONLY_TAPE_SOURCES],
                                        limit=db_bulk,
                                        older_than=datetime.datetime.utcnow(),
                                        activity=activity,
                                        process=process, total_processes=total_processes,
                                        thread=hb['assign_thread'], total_threads=hb['nr_threads'])
                record_timer('daemons.conveyor.finisher.000-get_next', (time.time() - ts) * 1000)
                if reqs:
                    logging.debug('%i:%i - updating %i requests for activity %s' % (process, hb['assign_thread'], len(reqs), activity))
                    sleeping = False

                for chunk in chunks(reqs, bulk):
                    try:
                        ts = time.time()
                        common.handle_requests(chunk)
                        record_timer('daemons.conveyor.finisher.handle_requests', (time.time() - ts) * 1000 / (len(chunk) if len(chunk) else 1))
                        record_counter('daemons.conveyor.finisher.handle_requests', len(chunk))
                    except:
                        logging.warn(str(traceback.format_exc()))
                if reqs:
                    logging.debug('%i:%i - finish to update %s finished requests for activity %s' % (process, hb['assign_thread'], len(reqs), activity))

        except (DatabaseException, DatabaseError), error:
            if isinstance(error.args[0], tuple) and (re.match('.*ORA-00054.*', error.args[0][0]) or ('ERROR 1205 (HY000)' in error.args[0][0])):
                logging.warn("%i:%i - Lock detected when handling request - skipping: %s" % (process, hb['assign_thread'], str(error)))
            else:
                logging.error("%i:%i - %s" % (process, hb['assign_thread'], traceback.format_exc()))
            sleeping = False
        except:
            sleeping = False
            logging.critical("%i:%i - %s" % (process, hb['assign_thread'], traceback.format_exc()))

        if once:
            return

    logging.info('%i:%i - graceful stop requests' % (process, hb['assign_thread']))

    heartbeat.die(executable, hostname, pid, hb_thread)

    logging.info('%i:%i - graceful stop done' % (process, hb['assign_thread']))


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(once=False, process=0, total_processes=1, total_threads=1, sleep_time=60, activities=None, bulk=100, db_bulk=1000):
    """
    Starts up the conveyer threads.
    """

    if once:
        logging.info('executing one finisher iteration only')
        finisher(once=once, activities=activities, bulk=bulk, db_bulk=db_bulk)

    else:

        logging.info('starting finisher threads')
        threads = [threading.Thread(target=finisher, kwargs={'process': process,
                                                             'total_processes': total_processes,
                                                             'thread': i,
                                                             'total_threads': total_threads,
                                                             'sleep_time': sleep_time,
                                                             'activities': activities,
                                                             'db_bulk': db_bulk,
                                                             'bulk': bulk}) for i in xrange(0, total_threads)]

        [t.start() for t in threads]

        logging.info('waiting for interrupts')

        # Interruptible joins require a timeout.
        while len(threads) > 0:
            threads = [t.join(timeout=3.14) for t in threads if t and t.isAlive()]



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#              http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Thomas Beermann, <thomas.beermann@cern.ch>, 2013

from json import loads
from random import uniform
from time import time

from rucio.client.didclient import DIDClient
from rucio.core import monitor
from rucio.tests.emulation.ucemulator import UCEmulator


def random_pick(some_list, probabilities):
    x = uniform(0, 1)
    cumulative_probability = 0.0
    for item, item_probability in zip(some_list, probabilities):
        cumulative_probability += item_probability
        if x < cumulative_probability:
            break
    return item


class UseCaseDefinition(UCEmulator):
    """
        Implements Jdoe use cases.
    """

    @UCEmulator.UseCase
    def LIST_DIDS_METADATA(self, scope, metadata):
        jdoe_account = 'jdoe'
        client = DIDClient(account=jdoe_account)

        print 'run with: ' + str(metadata)

        start = time()
        with monitor.record_timer_block('jdoe.list_dids_metadata'):
            dids = [did for did in client.list_dids(scope=scope, filters=metadata, type='dataset')]

        duration = time() - start
        cnt = len(dids)
        print 'got %d dids' % cnt

        monitor.record_counter('jdoe.list_dids_metadata.num_results', cnt)
        if cnt != 0:
            monitor.record_counter('jdoe.list_dids_metadata.time_per_did', duration / cnt)

        return {'no_datasets': cnt}

    def LIST_DIDS_METADATA_input(self, ctx):
        scope = random_pick(ctx.scopes, ctx.scope_probs)

        try:
            metadata = ctx.metadata[scope].next()
        except (StopIteration, ValueError):
            self.read_data(scope, ctx)
            metadata = ctx.metadata[scope].next()

        scope = metadata['project']
        metadata.pop('name', None)

        # print 'metadata picked: ' + str(metadata)

        return {'scope': scope,
                'metadata': metadata}

    def LIST_DIDS_WILDCARD(self, scope, wildcard):
        jdoe_account = 'jdoe'
        client = DIDClient(account=jdoe_account)

        print 'run with: ' + str(wildcard)
        start = time()
        with monitor.record_timer_block('jdoe.list_dids_wildcard'):
            dids = [did for did in client.list_dids(scope=scope, filters=wildcard, type='dataset')]

        duration = time() - start
        cnt = len(dids)
        print 'got %d dids' % cnt

        monitor.record_counter('jdoe.list_dids_wildcard.num_results', cnt)
        if cnt != 0:
            monitor.record_counter('jdoe.list_dids_wildcard.time_per_did', duration / cnt)

        return {'no_datasets': cnt}

    def LIST_DIDS_WILDCARD_input(self, ctx):
        scope = random_pick(ctx.scopes, ctx.scope_probs)

        try:
            metadata = ctx.metadata[scope].next()
        except (StopIteration, ValueError):
            self.read_data(scope, ctx)
            metadata = ctx.metadata[scope].next()

        scope = metadata['project']
        wildcard = {}
        wildcard['name'] = metadata['name']

        # print 'wildcard picked: ' + str(wildcard)

        return {'scope': scope,
                'wildcard': wildcard}

    def read_data(self, scope, ctx):
        input_file = open(ctx.input_files[scope], 'r')
        ctx.metadata[scope] = (loads(line.strip()) for line in input_file)

    def setup(self, ctx):
        ctx.metadata = {}
        print ctx
        for scope in ctx.scopes:
            self.read_data(scope, ctx)

        print 'Jdoe emulation starting'



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Thomas Beermann, <thomas.beermann@cern.ch>, 2016


from collections import deque
from threading import Lock, Timer


class ExpiringList:
        """
        Simple list with time based element expiration
        """

        def __init__(self, timeout=1):
            self._lock = Lock()
            self._timeout = timeout
            self._items = deque()

        def add(self, item):
            """Add event time
            """
            with self._lock:
                self._items.append(item)
                Timer(self._timeout, self._expire).start()

        def __len__(self):
            """
            Return number of active events
            """
            with self._lock:
                return len(self._timer)

        def _expire(self):
            """
            Remove any expired events
            """
            with self._lock:
                self._items.popleft()

        def to_set(self):
            return set(self._items)

        def __str__(self):
            with self._lock:
                return str(self._items)



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Luis Rodrigues, <luis.rodrigues@cern.ch>, 2013


# add_replica(rse='MOCK', scope='test', name='testfile0001', bytes=1, issuer='root')
# add_did(scope='test', name='testdataset1', type='DATASET', issuer='root')
# bulk: see add_replicas
# attach_dids(scope='ftest_scope', name='testdataset1', attachment={'dids': [{'scope': 'ftest_scope', 'name': 'testfile0002', 'type': 'FILE', 'bytes': 1 }, {'scope': 'ftest_scope', 'name': 'testfile0003', 'type': 'FILE', 'bytes': 1 }]}, issuer='root')
# add_replication_rule(dids=[{'scope': 'mock', 'name': 'testfile0001'}], account='root', copies=1, rse_expression='MOCK', grouping='NONE', weight=None, lifetime=None, issuer='root', locked=None, subscription_id=None)

commands = [
    # create the files
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0001', 'bytes': 'RANDOM:int:bytes1', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0002', 'bytes': 'RANDOM:int:bytes2', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0003', 'bytes': 'RANDOM:int:bytes3', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0004', 'bytes': 'RANDOM:int:bytes4', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0005', 'bytes': 'RANDOM:int:bytes5', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0006', 'bytes': 'RANDOM:int:bytes6', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0007', 'bytes': 'RANDOM:int:bytes7', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0008', 'bytes': 'RANDOM:int:bytes8', 'account': 'VAR:ACCOUNT'}},
    {'cmd': 'rse.add_replica', 'kwargs': {'rse': 'MOCK', 'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0009', 'bytes': 'RANDOM:int:bytes9', 'account': 'VAR:ACCOUNT'}},

    # create dataset
    {'cmd': 'did.add_did', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'type': 'DATASET', 'account': 'VAR:ACCOUNT'}},

    # parallel execute
    {'cmd': 'PARALLEL', 'list': [
        [
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0001', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes1'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0002', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes2'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0003', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes3'}], 'account': 'VAR:ACCOUNT'}},
        ],
        [
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0004', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes4'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0005', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes5'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0006', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes6'}], 'account': 'VAR:ACCOUNT'}},
        ],
        [
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0007', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes7'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0008', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes8'}], 'account': 'VAR:ACCOUNT'}},
            {'cmd': 'did.attach_dids', 'kwargs': {'scope': 'VAR:SCOPE', 'name': 'BLOCK:testdataset1', 'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0009', 'type': 'FILE', 'bytes': 'RANDOM:int:bytes9'}], 'account': 'VAR:ACCOUNT'}},
        ], ]
     },

    # add replication rules
    {'return': 'RULE1', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0001'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE2', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0002'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE3', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0003'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE4', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0004'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE5', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0005'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE6', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0006'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE7', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0007'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE8', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0008'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
    {'return': 'RULE9', 'cmd': 'rule.add_rule', 'kwargs': {'dids': [{'scope': 'VAR:SCOPE', 'name': 'BLOCK:testfile0009'}], 'account': 'VAR:ACCOUNT', 'copies': 1,
                                                           'rse_expression': 'MOCK', 'grouping': 'NONE', 'weight': None, 'lifetime': None, 'account': 'VAR:ACCOUNT', 'locked': None, 'subscription_id': None}},
]



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Angelos Molfetas, <angelos.molfetas@cern.ch>, 2011
# - Vincent Garonne,  <vincent.garonne@cern.ch> , 2011


def is_service_available(service_id):
    """
    Checks If a service is available.

    :param service_id: The service identifier.

    :returns: True if the service is available, otherwise False
    """
    raise NotImplementedError


def add_service_downtime(service_id, start_date=None, end_date=None, author=None, comment=None):
    """
    Schedules downtime for a specified service.

    :param service_id: The service identifier.
    :param start_date: The downtime start time.
    :param end_date: The downtime end time.
    :param contact: Service/shifter who manages the downtime.
    :param comment: Comment describing the reason of the service downtime.

    :returns: downtime_id
    """
    raise NotImplementedError


def delete_service_downtime(service_id, downtime_id):
    """
    Deletes a service downtime.

    :param service_id: The service identifier.
    :param downtime_id: The downtime identifier.

    :raises exception.NotFound if service_id/downtime_id is not found

    :returns: True if the service is available, otherwise False
    """
    raise NotImplementedError


def get_service_downtimes(service_id):
    """
    Returns a list of service downtimes for a service.

    :param service_id: The service identifier.

    :raises exception.NotFound if service_id is not found

    :retval Tuple containing (downtime_id,  start_date, end_date, author, comment)
    """
    raise NotImplementedError



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 2012-2014
# - Mario Lassnig, <mario.lassnig@cern.ch>, 2013-2015
# - Cedric Serfon, <cedric.serfon@cern.ch>, 2013-2015
# - Wen Guan, <wen.guan@cern.ch>, 2014-2015

"""
Conveyor is a daemon to manage file transfers.
"""

import json
import logging
import os
import socket
import sys
import threading
import time
import traceback

from collections import defaultdict
from ConfigParser import NoOptionError
from threadpool import ThreadPool, makeRequests

from rucio.common.config import config_get
from rucio.core import heartbeat
from rucio.core.monitor import record_counter, record_timer

from rucio.daemons.conveyor.submitter_utils import get_rses, get_transfers_from_requests, bulk_group_transfer, submit_transfer

logging.basicConfig(stream=sys.stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')

graceful_stop = threading.Event()


def submitter(once=False, rses=[],
              process=0, total_processes=1, thread=0, total_threads=1,
              mock=False, bulk=100, group_bulk=1, group_policy='rule', fts_source_strategy='auto',
              activities=None, activity_shares=None, sleep_time=600, max_sources=4):
    """
    Main loop to submit a new transfer primitive to a transfertool.
    """

    logging.info('Bulk submitter starting - process (%i/%i) thread (%i/%i)' % (process,
                                                                               total_processes,
                                                                               thread,
                                                                               total_threads))
    try:
        scheme = config_get('conveyor', 'scheme')
    except NoOptionError:
        scheme = None

    executable = ' '.join(sys.argv)
    hostname = socket.getfqdn()
    pid = os.getpid()
    hb_thread = threading.current_thread()

    logging.info('Bulk submitter started - process (%i/%i) thread (%i/%i)' % (process,
                                                                              total_processes,
                                                                              thread,
                                                                              total_threads))

    threadPool = ThreadPool(total_threads)
    activity_next_exe_time = defaultdict(time.time)
    while not graceful_stop.is_set():

        heartbeat.live(executable, hostname, pid, hb_thread)

        try:

            if activities is None:
                activities = [None]
            if rses:
                rse_ids = [rse['id'] for rse in rses]
            else:
                rse_ids = None

            for activity in activities:
                if activity_next_exe_time[activity] > time.time():
                    time.sleep(1)
                    continue

                logging.info("%s:%s Starting to submit jobs on activity: %s" % (process, thread, activity))

                logging.info("%s:%s Starting to get transfers" % (process, thread))
                ts = time.time()
                transfers = get_transfers_from_requests(process=process, total_processes=total_processes, thread=0, total_threads=1, rse_ids=rse_ids, mock=mock, bulk=bulk,
                                                        activity=activity, activity_shares=activity_shares, scheme=scheme, max_sources=max_sources)
                record_timer('daemons.conveyor.bulk_submitter.get_transfers_from_requests.per_transfer', (time.time() - ts) * 1000 / (len(transfers) if len(transfers) else 1))
                record_counter('daemons.conveyor.bulk_submitter.get_transfers_from_requests', len(transfers))
                record_timer('daemons.conveyor.bulk_submitter.get_transfers_from_requests.transfers', len(transfers))

                # group transfers
                logging.info("%s:%s Starting to group transfers" % (process, thread))
                ts = time.time()
                grouped_jobs = bulk_group_transfer(transfers, group_policy, group_bulk, fts_source_strategy)
                record_timer('daemons.conveyor.bulk_submitter.bulk_group_transfer', (time.time() - ts) * 1000 / (len(transfers) if len(transfers) else 1))

                logging.info("%s:%s Starting to submit transfers" % (process, thread))
                for external_host in grouped_jobs:
                    for job in grouped_jobs[external_host]:
                        # submit transfers
                        # job_requests = makeRequests(submit_transfer, args_list=[((external_host, job, 'bulk_submitter', process, thread), {})])
                        job_requests = makeRequests(submit_transfer, args_list=[((), {'external_host': external_host, 'job': job, 'submitter': 'bulk_submitter', 'process': process, 'thread': thread})])
                        [threadPool.putRequest(job_req) for job_req in job_requests]
                threadPool.wait()

                if len(transfers) < group_bulk:
                    logging.info('%i:%i - only %s transfers which is less than group bulk %s, sleep %s seconds' % (process, thread, len(transfers), group_bulk, sleep_time))
                    if activity_next_exe_time[activity] < time.time():
                        activity_next_exe_time[activity] = time.time() + sleep_time
        except:
            logging.critical('%s:%s %s' % (process, thread, traceback.format_exc()))

        if once:
            return

    logging.info('graceful stop requested')

    heartbeat.die(executable, hostname, pid, hb_thread)

    logging.info('graceful stop done')


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """

    graceful_stop.set()


def run(once=False,
        process=0, total_processes=1, total_threads=1, group_bulk=1, group_policy='rule',
        mock=False, rses=[], include_rses=None, exclude_rses=None, bulk=100, fts_source_strategy='auto',
        activities=[], activity_shares=None, sleep_time=600, max_sources=4):
    """
    Starts up the conveyer threads.
    """

    if mock:
        logging.info('mock source replicas: enabled')

    working_rses = None
    if rses or include_rses or exclude_rses:
        working_rses = get_rses(rses, include_rses, exclude_rses)
        logging.info("RSE selection: RSEs: %s, Include: %s, Exclude: %s" % (rses,
                                                                            include_rses,
                                                                            exclude_rses))
    else:
        logging.info("RSE selection: automatic")

    if activity_shares:

        try:
            activity_shares = json.loads(activity_shares)
        except:
            logging.critical('activity share is not a valid JSON dictionary')
            return

        try:
            if round(sum(activity_shares.values()), 2) != 1:
                logging.critical('activity shares do not sum up to 1, got %s - aborting' % round(sum(activity_shares.values()), 2))
                return
        except:
            logging.critical('activity shares are not numbers? - aborting')
            return

        activity_shares.update((share, int(percentage * bulk)) for share, percentage in activity_shares.items())
        logging.info('activity shares enabled: %s' % activity_shares)

    if once:
        logging.info('executing one submitter iteration only')
        submitter(once,
                  rses=working_rses,
                  mock=mock,
                  bulk=bulk,
                  group_bulk=group_bulk,
                  max_sources=max_sources,
                  group_policy=group_policy,
                  fts_source_strategy=fts_source_strategy,
                  activities=activities,
                  activity_shares=activity_shares)

    else:
        logging.info('starting submitter threads')
        threads = [threading.Thread(target=submitter, kwargs={'process': process,
                                                              'total_processes': total_processes,
                                                              'thread': 0,
                                                              'total_threads': total_threads,
                                                              'rses': working_rses,
                                                              'bulk': bulk,
                                                              'group_bulk': group_bulk,
                                                              'group_policy': group_policy,
                                                              'activities': activities,
                                                              'fts_source_strategy': fts_source_strategy,
                                                              'mock': mock,
                                                              'sleep_time': sleep_time,
                                                              'max_sources': max_sources,
                                                              'activity_shares': activity_shares})]

        [t.start() for t in threads]

        logging.info('waiting for interrupts')

        # Interruptible joins require a timeout.
        while len(threads) > 0:
            [t.join(timeout=3.14) for t in threads if t and t.isAlive()]



# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Vincent Garonne, <vincent.garonne@cern.ch>, 20125

"""
Conveyor is a daemon to manage file transfers.
"""

import logging
import sys
import threading
import time

from rucio.common.config import config_get
from rucio.daemons.conveyor.submitter_utils import get_transfer_requests_and_source_replicas, get_stagein_requests_and_source_replicas


logging.basicConfig(stream=sys.stdout,
                    level=getattr(logging, config_get('common', 'loglevel').upper()),
                    format='%(asctime)s\t%(process)d\t%(levelname)s\t%(message)s')

graceful_stop = threading.Event()


def submitter(once=False, process=0, total_processes=1, thread=0, total_threads=1):
    # print '%(thread)s/%(total_threads)s: Submitter ' % locals()
    s = time.time()
    requests, reqs_no_source, reqs_scheme_mismatch = get_transfer_requests_and_source_replicas(process=process, total_processes=total_processes, thread=thread, total_threads=total_threads)
    duration = time.time() - s
    nb_requests = len(requests)
    print '%(thread)s/%(total_threads)s: %(nb_requests)s xfers in %(duration)s seconds' % locals()
    for id in requests:
        print requests[id]
        break
    print reqs_no_source
    print reqs_scheme_mismatch

    s = time.time()
    requests, reqs_no_source = get_stagein_requests_and_source_replicas(process=process, total_processes=total_processes, thread=thread, total_threads=total_threads)
    duration = time.time() - s
    nb_requests = len(requests)
    print '%(thread)s/%(total_threads)s: %(nb_requests)s xfers in %(duration)s seconds' % locals()
    # Grouping per rule_id, dest_rse_id
    # Order by activity
    # jobs = requests_to_jobs(requests)
    jobs = requests

    # Then submit to FTS
    # for job in jobs:
    #   print jobs[job]
    #   fts_submit(job)
    #   update_request(job)

    # print '%(thread)s/%(total_threads)s: Stop ' % locals()
    return


def stop(signum=None, frame=None):
    """
    Graceful exit.
    """
    graceful_stop.set()


def run(once=False, total_threads=1):
    """
    Starts up the conveyer submitter threads.
    """
    print 'Starts up the conveyer submitter threads.'

    threads = []
    for i in xrange(total_threads):
        threads.append(threading.Thread(target=submitter, kwargs={'thread': i,
                                                                  'total_threads': total_threads,
                                                                  'once': once}))

    [t.start() for t in threads]

    logging.info('waiting for interrupts')
    while threads[0].is_alive():
        [t.join(timeout=3.14) for t in threads]
